import numpy as np
import scipy.sparse as sp

from angler.constants import DEFAULT_MATRIX_FORMAT


def createDws(w, s, dL, N, matrix_format=DEFAULT_MATRIX_FORMAT):
    # creates the derivative matrices
    # NOTE: python uses C ordering rather than Fortran ordering. Therefore the
    # derivative operators are constructed slightly differently than in MATLAB

    Nx = N[0]
    dx = dL[0]
    if len(N) is not 1:
        Ny = N[1]
        dy = dL[1]
    else:
        Ny = 1
        dy = 1
    if w is 'x':
        if Nx > 1:
            if s is 'f':
                dxf = sp.diags([-1, 1, 1], [0, 1, -Nx+1], shape=(Nx, Nx))
                Dws = 1/dx*sp.kron(dxf, sp.eye(Ny), format=matrix_format)
            else:
                dxb = sp.diags([1, -1, -1], [0, -1, Nx-1], shape=(Nx, Nx))
                Dws = 1/dx*sp.kron(dxb, sp.eye(Ny), format=matrix_format)
        else:
            Dws = sp.eye(Ny)            
    if w is 'y':
        if Ny > 1:
            if s is 'f':
                dyf = sp.diags([-1, 1, 1], [0, 1, -Ny+1], shape=(Ny, Ny))
                Dws = 1/dy*sp.kron(sp.eye(Nx), dyf, format=matrix_format)
            else:
                dyb = sp.diags([1, -1, -1], [0, -1, Ny-1], shape=(Ny, Ny))
                Dws = 1/dy*sp.kron(sp.eye(Nx), dyb, format=matrix_format)
        else:
            Dws = sp.eye(Nx)
    return Dws


def unpack_derivs(derivs):
    # takes derivs dictionary and returns tuple for convenience

    Dyb = derivs['Dyb']
    Dxb = derivs['Dxb']
    Dxf = derivs['Dxf']
    Dyf = derivs['Dyf']
    return (Dyb, Dxb, Dxf, Dyf)



from numpy import sqrt

EPSILON_0 = 8.85418782e-12
MU_0 = 1.25663706e-6
C_0 = sqrt(1/EPSILON_0/MU_0)
ETA_0 = sqrt(MU_0/EPSILON_0)

DEFAULT_MATRIX_FORMAT = 'csr'
DEFAULT_SOLVER = 'pardiso'
DEFAULT_LENGTH_SCALE = 1e-6  # microns

def n2_from_chi3(chi3):
    # takes chi3 in units of m^2 / V^2 and gives n2
    return 3*chi3/(C_0)/np.sqrt(eps_m)/(EPSILON_0)

    


from autograd import grad
import numpy as np

from angler.gradients import *

# maps components and nonlinearities to the corresponding grad function
GRADIENT_MAP = {
    'Ez': {
        'lin': grad_linear_Ez,
        'nl': grad_kerr_Ez},
    'Hx': {
        'lin': grad_linear_Hx,
        'nl': grad_kerr_Hx},
    'Hy': {
        'lin': grad_linear_Hy,
        'nl': grad_kerr_Hy},
    'Hz': {
        'lin': grad_linear_Hz,
        'nl': grad_kerr_Hz},
    'Ex': {
        'lin': grad_linear_Ex,
        'nl': grad_kerr_Ex},
    'Ey': {
        'lin': grad_linear_Ey,
        'nl': grad_kerr_Ey},
}

class Objective():

    """ Stores the objective function and its arguments 
        Useful for automatically selecting adjoint problems
    """
    def __init__(self, J, arg_list):        

        # if arg_list is just a single arg, create a list for compatibility
        if not isinstance(arg_list, list):
            arg_list = [arg_list]
        self.arg_list = arg_list
        
        # make sure number of arguments is consistent between J and arglist
        self.J = J
        sig = signature(J)
        num_args_J = len(sig.parameters.items())
        if (len(arg_list) != num_args_J):
            raise ValueError("number of arguemnts in J ({}) doesnt match that of arg_list ({})".format(len(arg_list), num_args_J))

    def is_linear(self):
        # is the objective function purely a function of linear fields?
        for arg in self.arg_list:
            if arg.nl:
                return False
        return True

    @property
    def J(self):
        # objective function
        return self._J

    @J.setter
    def J(self, J_new):
        # when objective function is reset, re-solve for the partials
        self._J = J_new
        self._autograd_J()
        self._get_gradients()

    def _autograd_J(self):

        sig = signature(self._J)
        num_args = len(sig.parameters.items())

        # note: eventually want to check whether J has eps_nl argument, then switch between linear and nonlinear depending.
        dJ_list = []

        for arg_index in range(num_args):
            dJ_list.append(grad(self._J, arg_index))
        self.dJ_list = dJ_list

    def _get_gradients(self):
        # loads in the gradient functions from the arg_list
        grad_fn_list = []
        for arg in self.arg_list:
            grad_fn_list.append(arg.gradient)
        self.grad_fn_list = grad_fn_list


class obj_arg():
    """ Defines an argument to the objective function.
        Specifies name, field component, and whether field is nonlinear
        Selects the corresponding gradient for the problem
    """

    def __init__(self, name, component='Ez', nl=False):

        # give a name to this argument (ex: "Ez_lin")
        self.name = str(name)

        # get field component
        if not component in {'Ex','Ey','Ez','Hx','Hy','Hz'}:
            raise ValueError("component must be in ['Ex','Ey','Ez','Hx','Hy','Hz'], was given '{}'".format(component))
        self.component = component

        # whether this is nonlinear field
        self.nl = nl

        self._select_gradient()

    def _select_gradient(self):
        # selects the correct gradient function.  

        lin_nl_selector = 'lin' if not self.nl else 'nl'
        self.gradient = GRADIENT_MAP[self.component][lin_nl_selector]

        if self.gradient is None:
            raise ValueError("Couldn't find a gradient for argument '{}' defined in gradient.py.".format(self.name))

if __name__ == "__main__":

    # example

    # define the arguments to the objective function
    f1 = obj_arg('Ez_lin', component='Ez', nl=False)
    f2 = obj_arg('Hx_lin', component='Hz', nl=False)
    f3 = obj_arg('Ex_nl', component='Ex', nl=True)
    arg_list = [f1, f2, f3]

    # define the objective function
    import autograd.numpy as npa
    def J(f1, f2, f3):

        A = npa.sum(npa.abs(f1*f2))
        B = npa.sum(npa.abs(f2*f3))
        C = npa.abs(A - B)
        return -C

    objective = Objective(J, arg_list)

    Nx = 40
    Ny = 4
    F1 = np.random.random((Nx, Ny))
    F2 = np.ones((Nx, Ny))
    F3 = np.random.random((Nx, Ny))



import numpy as np
import scipy.sparse as sp
from copy import deepcopy

from angler.constants import *
from angler.linalg import *


class mode:

    def __init__(self, neff, direction_normal, center, width, scale, order=1):
        self.neff = neff
        self.direction_normal = direction_normal
        self.center = center
        self.width = width
        self.order = order
        self.scale = scale

    def setup_src(self, simulation, matrix_format=DEFAULT_MATRIX_FORMAT):
        # compute the input power here using an only waveguide simulation
        self.compute_normalization(simulation, matrix_format=matrix_format)

        # insert the mode into the waveguide
        self.insert_mode(simulation, simulation.src, matrix_format=matrix_format)

    def compute_normalization(self, simulation, matrix_format=DEFAULT_MATRIX_FORMAT):
        # creates a single waveguide simulation, solves the source, computes the power

        # get some information from the permittivity
        original_eps = simulation.eps_r
        (Nx, Ny) = original_eps.shape
        eps_max = np.max(np.abs(original_eps))
        norm_eps = np.ones((Nx, Ny))

        # make a new simulation and get a new probe center
        simulation_norm = deepcopy(simulation)
        new_center = list(self.center)

        # compute where the source and waveguide should be
        if self.direction_normal == "x":
            inds_y = original_eps[self.center[0], :] > 1
            norm_eps[:, inds_y] = eps_max
            new_center[0] = Nx - new_center[0]
        elif self.direction_normal == "y":
            inds_x = original_eps[:, self.center[1]] > 1
            norm_eps[inds_x, :] = eps_max
            new_center[1] = Ny - new_center[1]
        else:
            raise ValueError("The value of direction_normal is not x or y!")

        # reset the permittivity to be a straight waveguide, solve fields, compute power
        simulation_norm.eps_r = norm_eps
        self.insert_mode(simulation_norm, simulation_norm.src, matrix_format=matrix_format)
        simulation_norm.solve_fields()
        W_in = simulation_norm.flux_probe(self.direction_normal, new_center, self.width)

        # save this value in the original simulation
        simulation.W_in = W_in

    def insert_mode(self, simulation, destination, matrix_format=DEFAULT_MATRIX_FORMAT):
        EPSILON_0_ = EPSILON_0*simulation.L0
        MU_0_ = MU_0*simulation.L0

        # first extract the slice of the permittivity
        if self.direction_normal == "x":
            inds_x = [self.center[0], self.center[0]+1]
            inds_y = [int(self.center[1]-self.width/2), int(self.center[1]+self.width/2)]
        elif self.direction_normal == "y":
            inds_x = [int(self.center[0]-self.width/2), int(self.center[0]+self.width/2)]
            inds_y = [self.center[1], self.center[1]+1]
        else:
            raise ValueError("The value of direction_normal is not x or y!")

        eps_r = simulation.eps_r[inds_x[0]:inds_x[1], inds_y[0]:inds_y[1]]
        N = eps_r.size

        Dxb = createDws('x', 'b', [simulation.dl], [N], matrix_format=matrix_format)
        Dxf = createDws('x', 'f', [simulation.dl], [N], matrix_format=matrix_format)

        vector_eps = EPSILON_0_*eps_r.reshape((-1,))
        vector_eps_x = EPSILON_0_*grid_average(eps_r, 'x').reshape((-1,))
        T_eps = sp.spdiags(vector_eps, 0, N, N, format=matrix_format)
        T_epsxinv = sp.spdiags(vector_eps_x**(-1), 0, N, N, format=matrix_format)

        if simulation.pol == 'Ez':
            A = np.square(simulation.omega)*MU_0_*T_eps + Dxf.dot(Dxb)

        elif simulation.pol == 'Hz':
            A = np.square(simulation.omega)*MU_0_*T_eps + T_eps.dot(Dxf).dot(T_epsxinv).dot(Dxb)

        est_beta = simulation.omega*np.sqrt(MU_0_*EPSILON_0_)*self.neff
        (vals, vecs) = solver_eigs(A, self.order, guess_value=np.square(est_beta))

        if self.order == 1:
            src = vecs
        else:
            src = vecs[:, self.order-1]

        src *= self.scale

        if self.direction_normal == 'x':
            src = src.reshape((1, -1))
            destination[inds_x[0]:inds_x[1], inds_y[0]:inds_y[1]] = np.abs(src)*np.sign(np.real(src))
        else:
            src = src.reshape((-1, 1))
            destination[inds_x[0]:inds_x[1], inds_y[0]:inds_y[1]] = np.abs(src)*np.sign(np.real(src))



# used for setup.py
name = "angler"

__version__ = '0.0.14'

# import the main classes
from .optimization import Optimization
from .simulation import Simulation

# import the various utilities
from .constants import *
from .plot import *
from .structures import *
from .utils import *


import numpy as np
import autograd.numpy as npa
import matplotlib.pylab as plt
from functools import wraps
from autograd.scipy.signal import convolve

""" This is meant to be a place to write convenience functions and objects to
    help with the optimization
"""

def eps2rho_bar(eps, eps_m):
    return (eps - 1) / (eps_m - 1)

def rho_bar2eps(rho_bar, eps_m):
    return rho_bar * (eps_m - 1) + 1

def deps_drho_bar(eps, eps_m):
    return 1 / (eps_m - 1)

def rho_bar(rho, eta=0.5, beta=100):
    num = npa.tanh(beta*eta) + npa.tanh(beta*(rho - eta))
    den = npa.tanh(beta*eta) + npa.tanh(beta*(1 - eta))
    return num / den

def drho_bar_drho(rho, eta=0.5, beta=100):
    rho_bar = partial(rho_bar, eta=eta, beta=beta)
    grad_rho_bar = grad(rho_bar)
    grad_rho_bar = np.vectorize(grad_rho_bar)
    return grad_rho_bar(rho)


class Binarizer():
    """
    Warning: Experimental Feature!!

        Takes a normal objective function and adds a binarization component to it
        For example:

            binarizer = Binarizer(design_region, eps_m)

            J = lambda e, e_nl, eps: npa.sum(npa.abs(e)) + npa.sum(npa.abs(e_nl))
            J_binarized = binarizer.density(J)

    where now J_binarized has multiplied J by some binary dependence on eps defined below in the density method.
    This can also be done in a fancy way like

        binarizer = Binarizer(design_region, eps_m)

        @binarizer.density
        J = lambda e, e_nl, eps: npa.sum(npa.abs(e)) + npa.sum(npa.abs(e_nl))

    In this case, J is changed in place with the decorator @binarizer.density
    """

    def __init__(self, design_region, eps_m, exp_const=1):

        self.design_region = design_region
        self.eps_m = eps_m
        self.exp_const = exp_const

    def density(self, J):
        """ Multiplies objective function by the density of eps_pixels near boundaries"""

        def J_bin(eps):
            """ Gives a number between 0 and 1 incidating how close each eps_r
                pixel is to being binarized, used to multiply with J()
            """

            # material density in design region
            rho = (eps - 1) / (self.eps_m - 1) * self.design_region

            # number of cells in design region
            N = npa.sum(self.design_region)

            # gray level map
            M_nd = 4 * rho * (1 - rho)

            # gray level indicator
            M_nd_scalar = npa.sum(M_nd) / N

            # average over each cell
            return 1 - M_nd_scalar

        # note, this is the actual generator being returned
        # defines how to combine J_bin (binarization function) with original J
        def J_new(*args, **kwargs):

            eps = args[2]
            return J(*args) * J_bin(eps)

        return J_new

    def density_exp(self, J):
        """ Multiplies objective function by the density of eps_pixels near boundaries"""

        def J_bin(eps):
            """ Gives a number between 0 and 1 incidating how close each eps_r
                pixel is to being binarized, used to multiply with J()
            """

            # material density in design region
            rho = (eps - 1) / (self.eps_m - 1) * self.design_region

            # number of cells in design region
            N = npa.sum(self.design_region)

            # gray level map
            M_nd = 4 * rho * (1 - rho)

            # gray level indicator
            M_nd_scalar = npa.sum(M_nd) / N

            # average over each cell
            # return 1 - M_nd_scalar
            # average over each cell
            return npa.exp(-self.exp_const*npa.abs(M_nd_scalar))

        # note, this is the actual generator being returned
        # defines how to combine J_bin (binarization function) with original J
        def J_new(*args, **kwargs):

            eps = args[2]
            return J(*args) * J_bin(eps)

        return J_new

    def smoothness(self, J):
        """ Multiplies objective function by the measure of smoothness of permittivity distribution"""

        def J_bin(eps):
            """ Gives a number between 0 and 1 incidating how close each eps_r
                pixel is to being binarized, used to multiply with J()
            """
            # material density in design region
            rho = (eps - 1) / (self.eps_m - 1) * self.design_region

            # number of cells in design region
            N = npa.sum(self.design_region)

            # gray level map
            M_nd = 4 * rho * (1 - rho)

            (Nx, Ny) = eps.shape

            width = 1
            N_conv = min(Nx, Ny)
            k = np.zeros((N_conv))
            k[N_conv//2:N_conv//2 + width] = 1/width

            N_conv = 2
            k = np.zeros((N_conv))
            k[1] = 1

            penalty = 0.0
            for i in range(Nx):
                strip_i = M_nd[i,:]
                output_i = convolve(k, strip_i)
                # penalty = penalty + 4*npa.sum(output_i*(1-output_i))
                penalty = penalty + npa.sum(npa.abs(output_i))

            for j in range(Ny):
                strip_j = M_nd[:,j]
                output_j = convolve(k, strip_j.T)
                # penalty = penalty + 4*npa.sum(output_j*(1-output_j))
                penalty = penalty + npa.sum(npa.abs(output_j))

            # print(penalty, N)
            penalty = penalty / N / 2
            return 1 - penalty

        # note, this is the actual generator being returned
        # defines how to combine J_bin (binarization function) with original J
        def J_new(*args, **kwargs):

            eps = args[2]
            # print(J_bin(eps))

            return J(*args) * J_bin(eps)

        return J_new

    # Implement more below vvvv if you want



import numpy as np

from angler.linalg import *


class Nonlinearity:

    def __init__(self, chi, nl_region, nl_type='kerr', eps_scale=False, eps_max=None):

        self.chi = chi
        self.nl_region = nl_region
        self.nl_type = nl_type
        self.eps_scale = eps_scale
        self.eps_max = eps_max
        self.eps_nl = []
        self.dnl_de = []
        self.dnl_deps = []

        if self.nl_type == 'kerr':
            if self.eps_scale:
                if self.eps_max is None:
                    raise AssertionError("Must provide eps_max when eps_scale is True") 

                else:
                    kerr_nonlinearity     = lambda e, eps_r:3*chi*nl_region*np.square(np.abs(e))*((eps_r-1)/(eps_max - 1))
                    kerr_nl_de            = lambda e, eps_r:3*chi*nl_region*np.conj(e)*((eps_r-1)/(eps_max - 1))
                    kerr_nl_deps          = lambda e, eps_r:3*chi*nl_region*np.square(np.abs(e))*(1/(eps_max - 1))

            else:
                kerr_nonlinearity     = lambda e, eps_r:3*chi*nl_region*np.square(np.abs(e))
                kerr_nl_de            = lambda e, eps_r:3*chi*nl_region*np.conj(e)
                kerr_nl_deps          = lambda e, eps_r:0

            self.eps_nl = kerr_nonlinearity
            self.dnl_de = kerr_nl_de
            self.dnl_deps = kerr_nl_deps

        else:
            raise AssertionError("Only 'kerr' type nonlinearity is currently supported") 



import numpy as np
import scipy.sparse as sp
from copy import deepcopy

from angler.linalg import construct_A, solver_direct, grid_average
from angler.derivatives import unpack_derivs
from angler.nonlinear_solvers import born_solve, newton_solve, newton_krylov_solve
from angler.source.mode import mode
from angler.nonlinearity import Nonlinearity
from angler.constants import (DEFAULT_LENGTH_SCALE, DEFAULT_MATRIX_FORMAT,
                              DEFAULT_SOLVER, EPSILON_0, MU_0)
from angler.filter import eps2rho
from angler.plot import plt_base_eps, plt_base

class Simulation:

    def __init__(self, omega, eps_r, dl, NPML, pol, L0=DEFAULT_LENGTH_SCALE):
        # initializes Fdfd object

        self.L0 = L0
        self.omega = omega
        self.NPML = NPML
        self.pol = pol
        self.dl = dl

        self._check_inputs()

        grid_shape = eps_r.shape
        if len(grid_shape) == 1:
            grid_shape = (grid_shape[0], 1)
            eps_r = np.reshape(eps_r, grid_shape)

        (Nx, Ny) = grid_shape

        self.Nx = Nx
        self.Ny = Ny

        self.mu_r = np.ones((self.Nx, self.Ny))
        self.src = np.zeros((self.Nx, self.Ny), dtype=np.complex64)

        self.xrange = [0, float(self.Nx*self.dl)]
        self.yrange = [0, float(self.Ny*self.dl)]

        self.NPML = [int(n) for n in self.NPML]
        self.omega = float(self.omega)
        self.dl = float(dl)

        # construct the system matrix
        self.eps_r = eps_r

        self.modes = []
        self.nonlinearity = []
        self.eps_nl = np.zeros(eps_r.shape)
        self.dnl_de = np.zeros(eps_r.shape)
        self.dnl_deps = np.zeros(eps_r.shape)

    def setup_modes(self):
        # calculates
        for modei in self.modes:
            modei.setup_src(self)

    def add_mode(self, neff, direction_normal, center, width,
                 scale=1, order=1):
        # adds a mode definition to the simulation
        new_mode = mode(neff, direction_normal, center, width,
                        scale=scale, order=order)
        self.modes.append(new_mode)

    def compute_nl(self, e, matrix_format=DEFAULT_MATRIX_FORMAT):
        # evaluates the nonlinear functions for a field e
        self.eps_nl = np.zeros(self.eps_r.shape)
        self.dnl_de = np.zeros(self.eps_r.shape)
        self.dnl_deps = np.zeros(self.eps_r.shape)
        for nli in self.nonlinearity:
            self.eps_nl = self.eps_nl + nli.eps_nl(e, self.eps_r)
            self.dnl_de = self.dnl_de + nli.dnl_de(e, self.eps_r)
            self.dnl_deps = self.dnl_deps + nli.dnl_deps(e, self.eps_r)
        Nbig = self.Nx*self.Ny
        Anl = sp.spdiags(self.omega**2*EPSILON_0*self.L0*self.eps_nl.reshape((-1,)), 0, Nbig, Nbig, format=matrix_format)
        self.Anl = Anl

    def add_nl(self, chi, nl_region, nl_type='kerr', eps_scale=False, eps_max=None):
        # adds a nonlinearity to the simulation
        new_nl = Nonlinearity(chi/np.square(self.L0), nl_region, nl_type, eps_scale, eps_max)
        self.nonlinearity.append(new_nl)

    @property
    def eps_r(self):
        return self.__eps_r

    @eps_r.setter
    def eps_r(self, new_eps):

        grid_shape = new_eps.shape
        if len(grid_shape) == 1:
            grid_shape = (grid_shape[0], 1)
            new_eps.reshape(grid_shape)

        (Nx, Ny) = grid_shape
        self.Nx = Nx
        self.Ny = Ny

        self.__eps_r = new_eps
        (A, derivs) = construct_A(self.omega, self.xrange, self.yrange,
                                  self.eps_r, self.NPML, self.pol, self.L0,
                                  matrix_format=DEFAULT_MATRIX_FORMAT,
                                  timing=False)
        self.A = A
        self.derivs = derivs
        self.fields = {f: None for f in ['Ex', 'Ey', 'Ez', 'Hx', 'Hy', 'Hz']}
        self.fields_nl = {f: None for f in ['Ex', 'Ey', 'Ez', 'Hx', 'Hy', 'Hz']}

    def solve_fields(self, include_nl=False, timing=False, averaging=False,
                     matrix_format=DEFAULT_MATRIX_FORMAT):
        # performs direct solve for A given source

        EPSILON_0_ = EPSILON_0*self.L0
        MU_0_ = MU_0*self.L0

        if include_nl==False:
            eps_tot = self.eps_r
            X = solver_direct(self.A, self.src*1j*self.omega, timing=timing)
        else:
            eps_tot = self.eps_r + self.eps_nl
            X = solver_direct(self.A + self.Anl, self.src*1j*self.omega, timing=timing)


        (Nx, Ny) = self.src.shape
        M = Nx*Ny
        (Dyb, Dxb, Dxf, Dyf) = unpack_derivs(self.derivs)

        if self.pol == 'Hz':
            if averaging:
                eps_x = grid_average(EPSILON_0_*(eps_tot), 'x')
                vector_eps_x = eps_x.reshape((-1,))
                eps_y = grid_average(EPSILON_0_*(eps_tot), 'y')
                vector_eps_y = eps_y.reshape((-1,))
            else:
                vector_eps_x = EPSILON_0_*(eps_tot).reshape((-1,))
                vector_eps_y = EPSILON_0_*(eps_tot).reshape((-1,))

            T_eps_x_inv = sp.spdiags(1/vector_eps_x, 0, M, M,
                                  format=matrix_format)
            T_eps_y_inv = sp.spdiags(1/vector_eps_y, 0, M, M,
                                  format=matrix_format)

            ex =  1/1j/self.omega * T_eps_y_inv.dot(Dyb).dot(X)
            ey = -1/1j/self.omega * T_eps_x_inv.dot(Dxb).dot(X)

            Ex = ex.reshape((Nx, Ny))
            Ey = ey.reshape((Nx, Ny))
            Hz = X.reshape((Nx, Ny))

            if include_nl==False:
                self.fields['Ex'] = Ex
                self.fields['Ey'] = Ey
                self.fields['Hz'] = Hz

            return (Ex, Ey, Hz)

        elif self.pol == 'Ez':
            hx = -1/1j/self.omega/MU_0_ * Dyb.dot(X)
            hy = 1/1j/self.omega/MU_0_ * Dxb.dot(X)

            Hx = hx.reshape((Nx, Ny))
            Hy = hy.reshape((Nx, Ny))
            Ez = X.reshape((Nx, Ny))

            if include_nl==False:
                self.fields['Hx'] = Hx
                self.fields['Hy'] = Hy
                self.fields['Ez'] = Ez

            return (Hx, Hy, Ez)

        else:
            raise ValueError('Invalid polarization: {}'.format(str(self.pol)))

    def solve_fields_nl(self,
                        timing=False, averaging=False,
                        Estart=None, solver_nl='hybrid', conv_threshold=1e-10,
                        max_num_iter=50,
                        matrix_format=DEFAULT_MATRIX_FORMAT):
        # solves for the nonlinear fields of the simulation.

        # note: F just stands for the field component (could be H or E depending on polarization)
        if solver_nl == 'born':
            (Fx, Fy, Fz, conv_array) = born_solve(self, Estart,
                                                  conv_threshold,
                                                  max_num_iter,
                                                  averaging=averaging)
        elif solver_nl == 'newton':
            (Fx, Fy, Fz, conv_array) = newton_solve(self, Estart,
                                                    conv_threshold,
                                                    max_num_iter,
                                                    averaging=averaging)
        elif solver_nl == 'LM':
            (Fx, Fy, Fz, conv_array) = LM_solve(self, Estart,
                                                conv_threshold,
                                                max_num_iter,
                                                averaging=averaging)
        elif solver_nl == 'krylov':
            (Fx, Fy, Fz, conv_array) = newton_krylov_solve(self, Estart,
                                                conv_threshold,
                                                max_num_iter,
                                                averaging=averaging)
        elif solver_nl == 'hybrid':
            (Fx, Fy, Fz, conv_array) = born_solve(self, Estart,
                                              conv_threshold, 20,
                                              averaging=averaging) 
            if conv_array[-1] > 1e-10:
                (Fx, Fy, Fz, conv_array) = newton_solve(self, Fz,
                                                    conv_threshold, 20,
                                                    averaging=averaging)                                       
        else:
            raise AssertionError("solver must be one of "
                                 "{'born', 'newton', 'LM'}")

            # return final nonlinear fields and an array of the convergences

        if self.pol == 'Ez':

            self.fields_nl['Hx'] = Fx
            self.fields_nl['Hy'] = Fy
            self.fields_nl['Ez'] = Fz

        elif self.pol == 'Hz':

            self.fields_nl['Ex'] = Fx
            self.fields_nl['Ey'] = Fy
            self.fields_nl['Hz'] = Fz

        else:
            raise ValueError('Invalid polarization: {}'.format(str(self.pol)))

        return (Fx, Fy, Fz, conv_array)

    def _check_inputs(self):
        # checks the inputs and makes sure they are kosher

        assert self.L0 > 0, "L0 must be a positive number, was supplied {},".format(str(self.L0))
        assert self.omega > 0, "omega must be a positive number, was supplied {},".format(str(self.omega))
        assert isinstance(self.NPML, list), "NPML must be a list of integers"
        assert len(self.NPML) == 2, "yrange must be a list of length 2, was supplied {}, which is of length {}".format(str(self.NPML), len(self.NPML))
        assert self.NPML[0] >= 0 and self.NPML[1] >= 0, "both elements of NPML must be >= 0"
        assert self.pol in ['Ez', 'Hz'], "pol must be one of 'Ez' or 'Hz'"
        # to do, check for correct types as well.

    def flux_probe(self, direction_normal, center, width, nl=False):
        # computes the total flux across the plane (line in 2D) defined by direction_normal, center, width

        # first extract the slice of the permittivity
        if direction_normal == "x":
            inds_x = [center[0], center[0]+1]
            inds_y = [int(center[1]-width/2), int(center[1]+width/2)]
        elif direction_normal == "y":
            inds_x = [int(center[0]-width/2), int(center[0]+width/2)]
            inds_y = [center[1], center[1]+1]
        else:
            raise ValueError("The value of direction_normal is neither x nor y!")

        if self.pol == 'Ez':

            if nl:
                field_val_Ez = self.fields_nl['Ez']
                field_val_Hy = self.fields_nl['Hy']
                field_val_Hx = self.fields_nl['Hx']
            else:
                field_val_Ez = self.fields['Ez']
                field_val_Hy = self.fields['Hy']
                field_val_Hx = self.fields['Hx']

            Ez_x = grid_average(field_val_Ez[inds_x[0]:inds_x[1]+1, inds_y[0]:inds_y[1]+1], 'x')[:-1,:-1]
            Ez_y = grid_average(field_val_Ez[inds_x[0]:inds_x[1]+1, inds_y[0]:inds_y[1]+1], 'y')[:-1,:-1]
            # NOTE: Last part drops the extra rows/cols used for grid_average

            if direction_normal == "x":
                Sx = -1/2*np.real(Ez_x*np.conj(field_val_Hy[inds_x[0]:inds_x[1], inds_y[0]:inds_y[1]]))
                return self.dl*np.sum(Sx)
            elif direction_normal == "y":
                Sy = 1/2*np.real(Ez_y*np.conj(field_val_Hx[inds_x[0]:inds_x[1], inds_y[0]:inds_y[1]]))
                return self.dl*np.sum(Sy)

        elif self.pol == 'Hz':

            if nl:
                field_val_Hz = self.fields_nl['Hz']
                field_val_Ey = self.fields_nl['Ey']
                field_val_Ex = self.fields_nl['Ex']
            else:
                field_val_Hz = self.fields['Hz']
                field_val_Ey = self.fields['Ey']
                field_val_Ex = self.fields['Ex']

            Hz_x = grid_average(field_val_Hz[inds_x[0]:inds_x[1]+1, inds_y[0]:inds_y[1]+1], 'x')[:-1, :-1]
            Hz_y = grid_average(field_val_Hz[inds_x[0]:inds_x[1]+1, inds_y[0]:inds_y[1]+1], 'y')[:-1, :-1]
            # NOTE: Last part drops the extra rows/cols used for grid_average

            if direction_normal == "x":
                Sx = 1/2*np.real(field_val_Ey[inds_x[0]:inds_x[1], inds_y[0]:inds_y[1]]*np.conj(Hz_x))
                return self.dl*np.sum(Sx)
            elif direction_normal == "y":
                Sy = -1/2*np.real(field_val_Ex[inds_x[0]:inds_x[1], inds_y[0]:inds_y[1]]*np.conj(Hz_y))
                return self.dl*np.sum(Sy)

    def init_design_region(self, design_region, eps_m, style=''):
        """ Initializes the design_region permittivity depending on style"""

        if style == 'full':
            # eps_m filled in design region
            eps_full = eps_m * np.ones(self.eps_r.shape)
            eps_full[design_region == 0] = self.eps_r[design_region == 0]
            self.eps_r = eps_full

        elif style == 'halfway':
            # halfway between 1 and eps_m in design region
            eps_halfway = self.eps_r
            eps_halfway[design_region == 1] = eps_m/2 + 1/2
            self.eps_r = eps_halfway

        elif style == 'empty':
            # nothing in design region
            eps_empty = np.ones(self.eps_r.shape)
            eps_empty[design_region == 0] = self.eps_r[design_region == 0]
            self.eps_r = eps_empty

        elif style == 'random':
            # random pixels in design region
            eps_random = (eps_m-1)*np.random.random(self.eps_r.shape)+1
            eps_random[design_region == 0] = self.eps_r[design_region == 0]
            self.eps_r = eps_random

        elif style == 'random_sym':
            # random pixels in design region.  Symmetric across y=0
            R = np.random.random(self.eps_r.shape)
            R_sym = 1/2*(np.fliplr(R) + R)
            eps_random = (eps_m-1)*R_sym+1
            eps_random[design_region == 0] = self.eps_r[design_region == 0]
            self.eps_r = eps_random

        self.rho = eps2rho(self.eps_r, eps_m)

    def compute_index_shift(self):
        """ Computes array of nonlinear refractive index shift"""

        _ = self.solve_fields()
        _ = self.solve_fields_nl()
        index_nl = np.sqrt(np.real(self.eps_r + self.eps_nl))
        index_lin = np.sqrt(np.real(self.eps_r))
        return np.abs(index_nl - index_lin)

    def plt_abs(self, nl=False, cbar=True, outline=True, ax=None, vmin=None, vmax=None, logscale=False, tiled_y=1):
        # plot np.absolute value of primary field (e.g. Ez/Hz)

        if self.fields[self.pol] is None:
            raise ValueError("need to solve the simulation first")

        eps_r = deepcopy(self.eps_r)
        eps_r = np.hstack(tiled_y*[eps_r])

        if nl:
            field_val = np.abs(self.fields_nl[self.pol])
        else:
            field_val = np.abs(self.fields[self.pol])

        field_val = np.hstack(tiled_y*[field_val])

        outline_val = np.abs(eps_r)
        
        if vmax is None:
            vmax = field_val.max()
        if vmin is None:
            vmin = 0.0

        cmap = "magma"

        return plt_base(field_val, outline_val, cmap, vmin, vmax, self.pol,
                        cbar=cbar, outline=outline, ax=ax, logscale=logscale)

    def plt_re(self, nl=False, cbar=True, outline=True, ax=None, tiled_y=1):
        """ Plots the real part of primary field (e.g. Ez/Hz)"""

        eps_r = deepcopy(self.eps_r)
        eps_r = np.hstack(tiled_y*[eps_r])

        if self.fields[self.pol] is None:
            raise ValueError("need to solve the simulation first")

        if nl:
            field_val = np.real(self.fields_nl[self.pol])
        else:
            field_val = np.real(self.fields[self.pol])

        field_val = np.hstack(tiled_y*[field_val])

        outline_val = np.abs(eps_r)
        vmin = -np.abs(field_val).max()
        vmax = +np.abs(field_val).max()
        cmap = "RdBu"

        return plt_base(field_val, outline_val, cmap, vmin, vmax, self.pol,
                        cbar=cbar, outline=outline, ax=ax)

    def plt_diff(self, cbar=True, outline=True, ax=None, vmax=None, tiled_y=1,
                 normalize=True):
        """ Plots the difference between |E| and |E_nl|"""

        # get the outline value
        eps_r = deepcopy(self.eps_r)
        eps_r = np.hstack(tiled_y*[eps_r])
        outline_val = np.abs(eps_r)

        # get the fields and tile them
        field_lin = np.abs(self.fields[self.pol])
        field_lin = np.hstack(tiled_y*[field_lin])
        field_nl = np.abs(self.fields_nl[self.pol])
        field_nl = np.hstack(tiled_y*[field_nl])

        # take the difference, normalize by the max E_lin field if desired
        field_diff = field_lin - field_nl
        if normalize:
            field_diff = field_diff/np.abs(field_lin).max()

        # set limits
        if vmax is None:
            vmax = np.abs(field_diff).max()
        vmin = -vmax

        return plt_base(field_diff, outline_val, 'RdYlBu', vmin, vmax,
                        self.pol, cbar=cbar, outline=outline, ax=ax)

    def plt_eps(self, cbar=True, outline=True, ax=None, tiled_y=1):
        # plot the permittivity distribution

        eps_r = deepcopy(self.eps_r)
        eps_r = np.hstack(tiled_y*[eps_r])

        eps_val = np.abs(eps_r)
        outline_val = np.abs(eps_r)
        vmin = np.abs(eps_r).min()
        vmax = np.abs(eps_r).max()
        cmap = "Greys"

        return plt_base_eps(eps_val, outline_val, cmap, vmin, vmax, cbar=cbar,
                            outline=outline, ax=ax)



import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spl

try:
    from pyMKL import pardisoSolver
    SOLVER = 'pardiso'
except:
    SOLVER = 'scipy'

from time import time

from angler.constants import DEFAULT_MATRIX_FORMAT, DEFAULT_SOLVER
from angler.constants import EPSILON_0, MU_0
from angler.pml import S_create
from angler.derivatives import createDws


def grid_average(center_array, w):
    # computes values at cell edges

    xy = {'x': 0, 'y': 1}
    center_shifted = np.roll(center_array, 1, axis=xy[w])
    avg_array = (center_shifted+center_array)/2
    return avg_array


def dL(N, xrange, yrange=None):
    # solves for the grid spacing

    if yrange is None:
        L = np.array([np.diff(xrange)[0]])  # Simulation domain lengths
    else:
        L = np.array([np.diff(xrange)[0],
                      np.diff(yrange)[0]])  # Simulation domain lengths
    return L/N


def is_equal(matrix1, matrix2):
    # checks if two sparse matrices are equal

    return (matrix1 != matrix2).nnz == 0


def construct_A(omega, xrange, yrange, eps_r, NPML, pol, L0,
                averaging=True,
                timing=False,
                matrix_format=DEFAULT_MATRIX_FORMAT):
    # makes the A matrix
    N = np.asarray(eps_r.shape)  # Number of mesh cells
    M = np.prod(N)  # Number of unknowns

    EPSILON_0_ = EPSILON_0*L0
    MU_0_ = MU_0*L0

    if pol == 'Ez':
        vector_eps_z = EPSILON_0_*eps_r.reshape((-1,))
        T_eps_z = sp.spdiags(vector_eps_z, 0, M, M, format=matrix_format)

        (Sxf, Sxb, Syf, Syb) = S_create(omega, L0, N, NPML, xrange, yrange, matrix_format=matrix_format)

        # Construct derivate matrices
        Dyb = Syb.dot(createDws('y', 'b', dL(N, xrange, yrange), N, matrix_format=matrix_format))
        Dxb = Sxb.dot(createDws('x', 'b', dL(N, xrange, yrange), N, matrix_format=matrix_format))
        Dxf = Sxf.dot(createDws('x', 'f', dL(N, xrange, yrange), N, matrix_format=matrix_format))
        Dyf = Syf.dot(createDws('y', 'f', dL(N, xrange, yrange), N, matrix_format=matrix_format))

        A = (Dxf*1/MU_0_).dot(Dxb) \
            + (Dyf*1/MU_0_).dot(Dyb) \
            + omega**2*T_eps_z

    elif pol == 'Hz':
        if averaging:
            vector_eps_x = grid_average(EPSILON_0_*eps_r, 'x').reshape((-1,))
            vector_eps_y = grid_average(EPSILON_0_*eps_r, 'y').reshape((-1,))
        else:
            vector_eps_x = EPSILON_0_*eps_r.reshape((-1,))
            vector_eps_y = EPSILON_0_*eps_r.reshape((-1,))

        # Setup the T_eps_x, T_eps_y, T_eps_x_inv, and T_eps_y_inv matrices
        T_eps_x = sp.spdiags(vector_eps_x, 0, M, M, format=matrix_format)
        T_eps_y = sp.spdiags(vector_eps_y, 0, M, M, format=matrix_format)
        T_eps_x_inv = sp.spdiags(1/vector_eps_x, 0, M, M, format=matrix_format)
        T_eps_y_inv = sp.spdiags(1/vector_eps_y, 0, M, M, format=matrix_format)

        (Sxf, Sxb, Syf, Syb) = S_create(omega, L0, N, NPML, xrange, yrange, matrix_format=matrix_format)

        # Construct derivate matrices
        Dyb = Syb.dot(createDws('y', 'b', dL(N, xrange, yrange), N, matrix_format=matrix_format))
        Dxb = Sxb.dot(createDws('x', 'b', dL(N, xrange, yrange), N, matrix_format=matrix_format))
        Dxf = Sxf.dot(createDws('x', 'f', dL(N, xrange, yrange), N, matrix_format=matrix_format))
        Dyf = Syf.dot(createDws('y', 'f', dL(N, xrange, yrange), N, matrix_format=matrix_format))

        A =   Dxf.dot(T_eps_x_inv).dot(Dxb) \
            + Dyf.dot(T_eps_y_inv).dot(Dyb) \
            + omega**2*MU_0_*sp.eye(M)

    else:
        raise ValueError("something went wrong and pol is not one of Ez, Hz, instead was given {}".format(pol))

    derivs = {
        'Dyb' : Dyb,
        'Dxb' : Dxb,
        'Dxf' : Dxf,
        'Dyf' : Dyf
    }

    return (A, derivs)


def solver_eigs(A, Neigs, guess_value=0, guess_vector=None, timing=False):
    # solves for the eigenmodes of A

    if timing:
        start = time()
    (values, vectors) = spl.eigs(A, k=Neigs, sigma=guess_value, v0=guess_vector, which='LM')
    if timing:
        end = time()
        print('Elapsed time for eigs() is %.4f secs' % (end - start))
    return (values, vectors)


def solver_direct(A, b, timing=False, solver=SOLVER):
    # solves linear system of equations

    b = b.astype(np.complex128)
    b = b.reshape((-1,))

    if not b.any():
        return np.zeros(b.shape)

    if timing:
        t = time()

    if solver.lower() == 'pardiso':
        pSolve = pardisoSolver(A, mtype=13) # Matrix is complex unsymmetric due to SC-PML
        pSolve.factor()
        x = pSolve.solve(b)
        pSolve.clear()

    elif solver.lower() == 'scipy':
        x = spl.spsolve(A, b)

    else:
        raise ValueError('Invalid solver choice: {}, options are pardiso or scipy'.format(str(solver)))

    if timing:
        print('Linear system solve took {:.2f} seconds'.format(time()-t))

    return x


def solver_complex2real(A11, A12, b, timing=False, solver=SOLVER):
    # solves linear system of equations [A11, A12; A21*, A22*]*[x; x*] = [b; b*]

    b = b.astype(np.complex128)
    b = b.reshape((-1,))
    N = b.size

    if not b.any():
        return np.zeros(b.shape)

    b_re = np.real(b).astype(np.float64)
    b_im = np.imag(b).astype(np.float64)

    Areal = sp.vstack((sp.hstack((np.real(A11) + np.real(A12), - np.imag(A11) + np.imag(A12))),
                       sp.hstack((np.imag(A11) + np.imag(A12), np.real(A11) - np.real(A12)))))

    if timing:
        t = time()

    if solver.lower() == 'pardiso':
        pSolve = pardisoSolver(Areal, mtype=11)  # Matrix is real unsymmetric
        pSolve.factor()
        x = pSolve.solve(np.hstack((b_re, b_im)))
        pSolve.clear()

    elif solver.lower() == 'scipy':
        x = spsolve(Areal, np.hstack((b_re, b_im)))

    else:
        raise ValueError('Invalid solver choice: {}, options are pardiso or scipy'.format(str(solver)))

    if timing:
        print('Linear system solve took {:.2f} seconds'.format(time()-t))

    return (x[:N] + 1j*x[N:2*N])



import numpy as np
import scipy.sparse as sp
import copy
import progressbar
import matplotlib.pylab as plt
from scipy.optimize import minimize, fmin_l_bfgs_b
from autograd import grad

from angler.constants import *
from angler.filter import (eps2rho, rho2eps, get_W, deps_drhob, drhob_drhot,
                    drhot_drho, rho2rhot, drhot_drho, rhot2rhob)


class Optimization():

    def __init__(self, objective, simulation, design_region, eps_m=5,
                 R=None, eta=0.5, beta=1e-9,
                 field_start='linear', nl_solver='newton', max_ind_shift=None):

        # store essential objects
        self.objective = objective
        self.simulation = simulation
        self.design_region = design_region

        # store and compute filter and projection objects
        self.R = R
        (Nx, Ny) = self.simulation.eps_r.shape
        if self.R is not None:
            self.W = get_W(Nx, Ny, self.design_region,  NPML=self.simulation.NPML, R=self.R)
        else:
            self.W = sp.eye(Nx*Ny, dtype=np.complex64)
        self.eta = eta
        self.beta = beta

        # stre other optional parameters
        self.eps_m = eps_m
        self.field_start = field_start
        self.nl_solver = nl_solver
        self.max_ind_shift = max_ind_shift

        # store things that will be used later
        self.src_amplitudes = []
        self.objfn_list = []
        self.fields_current = False     # are the field_args current?  if not, will need to recompute

    def _solve_objfn_arg_fields(self, simulation):
        """ 
            solves for all of the fields needed in the objective function.
            also checks if it's linear (and if so, doesnt solve_nl)
        """

        # do something only if the fields are not current, or there are no stored fields
        if not self.fields_current or not self.field_arg_list:

            _ = simulation.solve_fields()
            if not self.objective.is_linear():
                _ = simulation.solve_fields_nl()

            # prepare a list of arguments that correspond to obj_fn arguments
            field_arg_list = []
            for arg in self.objective.arg_list:
                if not arg.nl:
                    field = simulation.fields[arg.component]
                else:
                    field = simulation.fields_nl[arg.component]
                if field is None:
                    raise ValueError("couldn't find a field defined for component '{}'.  Could be the wrong polarization (simulation is '{}' polarization).".format(arg.component, self.simulation.pol))
                field_arg_list.append(field)

            # store these arguments (so that they dont have to be recomputed)
            self.field_arg_list = field_arg_list

            # lets other functions know the field_arg_list is up to date
            self.fields_current = True


    def compute_J(self, simulation):
        """ Returns the current objective function of a simulation"""

        # stores all of the fields for the objective function in self.field_arg_list
        self._solve_objfn_arg_fields(simulation)

        # pass these arguments to the objective function
        return self.objective.J(*self.field_arg_list)

    def compute_dJ(self, simulation, design_region):
        """ Returns the current grad of a simulation"""

        # stores all of the fields for the objective function in self.field_arg_list
        self._solve_objfn_arg_fields(simulation)

        # sum up gradient contributions from each argument in J
        gradient_sum = 0
        for gradient_fn, dJ, arg in zip(self.objective.grad_fn_list, self.objective.dJ_list, self.objective.arg_list):
            if not arg.nl:
                Fz = simulation.fields[simulation.pol]
            else:
                Fz = simulation.fields_nl[simulation.pol]
            gradient = gradient_fn(self, dJ, Fz, self.field_arg_list)
            gradient_sum += gradient

        return gradient_sum

    def check_deriv(self, Npts=5, d_rho=1e-3):
        """ Returns a list of analytical and numerical derivatives to check grad accuracy"""

        self.fields_current = False
        self.simulation.eps_r = rho2eps(rho=self.simulation.rho, eps_m=self.eps_m, W=self.W,
                                        eta=self.eta, beta=self.beta)

        # solve for the linear fields and grad of the linear objective function
        grad_avm = self.compute_dJ(self.simulation, self.design_region)
        J_orig = self.compute_J(self.simulation)

        avm_grads = []
        num_grads = []

        # for a number of points
        for _ in range(Npts):

            # pick a random point within the design region
            x, y = np.where(self.design_region == 1)
            i = np.random.randint(len(x))
            pt = [x[i], y[i]]

            # create a new, perturbed permittivity
            rho_new = copy.deepcopy(self.simulation.rho)
            rho_new[pt[0], pt[1]] += d_rho

            # make a copy of the current simulation
            sim_new = copy.deepcopy(self.simulation)
            eps_new = rho2eps(rho=rho_new, eps_m=self.eps_m, W=self.W,
                              eta=self.eta, beta=self.beta)

            sim_new.rho = rho_new
            sim_new.eps_r = eps_new

            self.fields_current = False

            # solve for the fields with this new permittivity
            J_new = self.compute_J(sim_new)

            # compute the numerical grad
            grad_num = (J_new - J_orig) / d_rho

            # append both grads to lists
            avm_grads.append(grad_avm[pt[0], pt[1]])
            num_grads.append(grad_num)

        return avm_grads, num_grads

    def _make_progressbar(self, N):
        """ Returns a progressbar to use during optimization"""

        if self.max_ind_shift is not None:

            bar = progressbar.ProgressBar(widgets=[
                ' ', progressbar.DynamicMessage('ObjectiveFn'),
                ' ', progressbar.DynamicMessage('ObjectiveFn_Normalized'),
                ' Iteration: ',
                ' ', progressbar.Counter(), '/%d' % N,
                ' ', progressbar.AdaptiveETA(),
            ], max_value=N)

        else:

            bar = progressbar.ProgressBar(widgets=[
                ' ', progressbar.DynamicMessage('ObjectiveFn'),
                ' Iteration: ',
                ' ', progressbar.Counter(), '/%d' % N,
                ' ', progressbar.AdaptiveETA(),
            ], max_value=N)

        return bar

    def _update_progressbar(self, pbar, iteration, J):

        if self.max_ind_shift is not None:
            objfn_norm = J/np.max(np.square(np.abs(self.simulation.src)))
            pbar.update(iteration, ObjectiveFn=J, ObjectiveFn_Normalized=objfn_norm)
        else:
            pbar.update(iteration, ObjectiveFn=J)

    def run(self, method='LBFGS', Nsteps=100, step_size=0.1,
            beta1=0.9, beta2=0.999, verbose=True, temp_plt=None):
        """ Runs an optimization."""

        self.Nsteps = Nsteps
        self.verbose = verbose
        self.temp_plt = temp_plt

        self._check_temp_plt()

        # get the material density from the simulation if only the first time being run
        if self.simulation.rho is None:
            eps = copy.deepcopy(self.simulation.eps_r)
            self.simulation.rho = eps2rho(eps)
        
        self.fields_current = False
        
        allowed = ['LBFGS', 'GD', 'ADAM']

        if method.lower() in ['lbfgs']:
            self._run_LBFGS()

        elif method.lower() == 'gd':
            self._run_GD(step_size=step_size)

        elif method.lower() == 'adam':
            self._run_ADAM(step_size=step_size, beta1=beta1, beta2=beta2)

        else:
            raise ValueError("'method' must be in {}".format(allowed))

    def _run_GD(self, step_size):
        """ Performs simple grad descent optimization"""

        pbar = self._make_progressbar(self.Nsteps)

        for iteration in range(self.Nsteps):

            J = self.compute_J(self.simulation)
            self.objfn_list.append(J)

            self._update_progressbar(pbar, iteration, J)

            self._set_source_amplitude()

            grad = self.compute_dJ(self.simulation, self.design_region)

            if self.temp_plt is not None:
                if np.mod(iteration, self.temp_plt.it_plot) == 0:
                    self.plot_it(iteration)

            self._update_rho(grad, step_size)

    def _run_ADAM(self, step_size, beta1, beta2):
        """ Performs simple grad descent optimization"""
        pbar = self._make_progressbar(self.Nsteps)

        for iteration in range(self.Nsteps):

            J = self.compute_J(self.simulation)
            self.objfn_list.append(J)
            # pbar.update(iteration, ObjectiveFn=J)
            self._update_progressbar(pbar, iteration, J)

            self._set_source_amplitude()

            grad = self.compute_dJ(self.simulation, self.design_region)

            if iteration == 0:
                mopt = np.zeros(grad.shape)
                vopt = np.zeros(grad.shape)

            (grad_adam, mopt, vopt) = self._step_adam(grad, mopt, vopt, iteration, beta1, beta2,)

            if self.temp_plt is not None:
                if np.mod(iteration, self.temp_plt.it_plot) == 0:
                    self.plot_it(iteration)

            self._update_rho(grad_adam, step_size)

    def _run_LBFGS(self):
        """Performs L-BFGS Optimization of objective function w.r.t. eps_r"""

        pbar = self._make_progressbar(self.Nsteps)

        def _objfn(rho, *argv):
            """ Returns objective function given some permittivity distribution"""

            self._set_design_region(rho)
            J = self.compute_J(self.simulation)

            # return minus J because we technically will minimize
            return -J

        def _grad(rho,  *argv):
            """ Returns full grad given some permittivity distribution"""

            self._set_design_region(rho)

            # compute grad, extract design region, turn into vector, return
            grad = self.compute_dJ(self.simulation, self.design_region)
            grad_vec = self._get_design_region(grad)

            return -grad_vec

        # this simple callback function gets run each iteration
        # keeps track of the current iteration step for the progressbar
        # also resets eps on the simulation
        iter_list = [0]

        def _update_iter_count(x_current):
            J = self.compute_J(self.simulation)
            self._update_progressbar(pbar, iter_list[0], J)
            self.objfn_list.append(J)
            self._set_design_region(x_current)
            self._set_source_amplitude()
            if self.temp_plt is not None:
                if np.mod(iter_list[0], self.temp_plt.it_plot) == 0:
                    self.plot_it(iter_list[0])

            iter_list[0] += 1

        N_des = np.sum(self.design_region == 1)              # num points in design region
        rho_bounds = tuple([(0, 1) for _ in range(N_des)])   # bounds on rho {0, 1}

        # start eps off with the one currently within design region
        rho = copy.deepcopy(self.simulation.rho)
        rho0 = self._get_design_region(rho)
        rho0 = np.reshape(rho0, (-1,))

        # minimize
        (rho_final, _, _) = fmin_l_bfgs_b(_objfn, rho0, fprime=_grad, args=(), approx_grad=0,
                            bounds=rho_bounds, m=10, factr=10,
                            pgtol=1e-15, epsilon=1e-08, iprint=-1,
                            maxfun=15000, maxiter=self.Nsteps, disp=self.verbose,
                            callback=_update_iter_count, maxls=20)

        # finally, set the simulation permittivity to that found via optimization
        self._set_design_region(rho_final)

    def plot_it(self, iteration):
        Nplots = len(self.temp_plt.plot_what)
        plt.close('all')

        Ez_lin = self.simulation.fields['Ez']

        # This is what was used for the T-port in the paper    
        vmin = 4 * np.sqrt(self.simulation.W_in)
        vmax = np.abs(Ez_lin).max()/1.5

        # # This is more general but less flexible
        # if self.temp_plt.vlims[0] == None:
        #     vmin = np.abs(Ez_lin).min()
        # else:
        #     vmin = self.temp_plt.vlims[0]
        # if self.temp_plt.vlims[1] == None:
        #     vmax = np.abs(Ez_lin).max()
        # else:
        #     vmax = self.temp_plt.vlims[1]

        if Nplots == 4:
            f, faxs = plt.subplots(2, 2, figsize=self.temp_plt.figsize)
            axs = faxs.ravel()
        else:
            f, axs = plt.subplots(1, Nplots, figsize=self.temp_plt.figsize)

        for n, plots in enumerate(self.temp_plt.plot_what):
            if plots == 'eps':
                ax = axs[n]
                self.simulation.plt_eps(outline=False, cbar=False, ax=ax)
                ax.set_title('Permittivity')
                ax.annotate('Iteration:%4d' % np.int(iteration), (0.2, 0.92),
                    xycoords='axes fraction',
                    size='medium',
                    color='k',
                    horizontalalignment='center',
                    verticalalignment='center')
            if plots == 'of':
                ax = axs[n]
                self.plt_objs(ax=ax)
                ax.set_title('Objective')
            if plots == 'elin':
                ax = axs[n]
                self.simulation.plt_abs(outline=True, cbar=False, ax=ax, logscale=True, vmin=vmin, vmax=vmax)
                ax.set_title('Linear field')
            if plots == 'enl':
                ax = axs[n]
                self.simulation.plt_abs(outline=True, cbar=False, ax=ax, nl=True, logscale=True, vmin=vmin, vmax=vmax)
                ax.set_title('Nonlinear field')
        
        fname = self.temp_plt.folder + ('it%06d.png' % np.int(iteration))
        plt.savefig(fname, dpi=self.temp_plt.dpi)

    def _set_design_region(self, x):
        """ Inserts a vector x into design region of simulation.rho """

        rho_vec = np.reshape(copy.deepcopy(self.simulation.rho), (-1,))
        des_vec = np.reshape(self.design_region, (-1,))

        # Only update the rho if it actually differs from the current one
        # If it doesn't, we don't want to erase the stored fields

        if np.linalg.norm(x - rho_vec[des_vec == 1])/np.linalg.norm(x) > 1e-10:
            rho_vec[des_vec == 1] = x
            rho_new = np.reshape(rho_vec, self.simulation.rho.shape)
            self.simulation.rho = rho_new
            eps_new = rho2eps(rho=rho_new, eps_m=self.eps_m, W=self.W,
                              eta=self.eta, beta=self.beta)
            self.simulation.eps_r = eps_new

        self.fields_current = False

    def _get_design_region(self, spatial_array):
        """ Returns a vector of the elements of spatial_array that are in design_region"""

        spatial_vec = copy.deepcopy(np.ndarray.flatten(spatial_array))
        des_vec = np.ndarray.flatten(self.design_region)
        x = spatial_vec[des_vec == 1]
        return x

    def _set_source_amplitude(self, epsilon=1e-2, N=1):
        """ If max_index_shift specified, sets the self.simulation.src amplitude
            low enough so that this is satisfied.
            'epsilon' is the amount to subtract from source to get it under.
        """

        # keep a running list of the source amplitudes
        self.src_amplitudes.append(np.max(np.abs(self.simulation.src)))

        # if a max index shift is specified
        if self.max_ind_shift is not None:

            # for a number of iterations
            for _ in range(N):

                # compute the index shift and update the source according to the ratio
                dn = self.simulation.compute_index_shift()
                max_dn = np.max(dn)
                ratio = self.max_ind_shift / max_dn

                self.simulation.src = self.simulation.src * (np.sqrt(ratio) - epsilon)
        self.fields_current = False

    def _update_rho(self, grad, step_size):
        """ Manually updates the permittivity with the grad info """

        self.simulation.rho = self.simulation.rho + self.design_region * step_size * grad
        self.simulation.rho[self.simulation.rho < 0] = 0
        self.simulation.rho[self.simulation.rho > 1] = 1

        self.simulation.eps_r = rho2eps(self.simulation.rho, self.eps_m, self.W,
                                        eta=self.eta, beta=self.beta)
        self.fields_current = False

    def _step_adam(self, gradient, mopt_old, vopt_old, iteration, beta1, beta2, epsilon=1e-8):
        """ Performs one step of adam optimization"""

        mopt = beta1 * mopt_old + (1 - beta1) * gradient
        mopt_t = mopt / (1 - beta1**(iteration + 1))
        vopt = beta2 * vopt_old + (1 - beta2) * (np.square(gradient))
        vopt_t = vopt / (1 - beta2**(iteration + 1))
        grad_adam = mopt_t / (np.sqrt(vopt_t) + epsilon)

        return (grad_adam, mopt, vopt)

    def plt_objs(self, norm=None, ax=None):
        """ Plots objective function vs. iteration"""

        iter_range = range(1, len(self.objfn_list) + 1)
        if norm == 'field':
            obj_scaled = [o/a for o, a in zip(self.objfn_list, self.src_amplitudes)]
            ax.set_ylabel('objective function / field')
        elif norm == 'power':
            obj_scaled = [o/a**2 for o, a in zip(self.objfn_list, self.src_amplitudes)]
            ax.set_ylabel('objective function / power')
        else:
            obj_scaled = self.objfn_list
            ax.set_ylabel('objective function')


        ax.plot(iter_range,  obj_scaled)
        ax.set_xlabel('iteration number')
        ax.set_title('optimization results')
        return ax

    def scan_frequency(self, Nf=50, df=1/20, pbar=True):
        """ Scans the objective function vs. frequency """

        # create frequencies (in Hz)
        delta_f = self.simulation.omega*df
        freqs = 1/2/np.pi*np.linspace(self.simulation.omega - delta_f/2,
                                      self.simulation.omega + delta_f/2,  Nf)

        if pbar:
            bar = progressbar.ProgressBar(max_value=Nf)

        # loop through frequencies
        objs = []
        for i, f in enumerate(freqs):

            if pbar:
                bar.update(i + 1)

            # make a new simulation object
            sim_new = copy.deepcopy(self.simulation)

            # reset the simulation to compute new A (hacky way of doing it)
            sim_new.omega = 2*np.pi*f
            sim_new.eps_r = self.simulation.eps_r

            self.fields_current = False

            # compute objective function and append to list
            obj_fn = self.compute_J(sim_new)
            objs.append(obj_fn)

        # compute HM
        objs_array = np.array(objs)
        HM = np.max(objs_array)/2
        above_HM = objs_array > HM

        # does a scan up and down from the midpoint and counts number above HM in this peak
        num_above_HM = 0
        for i in range(int(Nf/2), Nf):
            if not above_HM[i]:
                break
            num_above_HM += 1
        for i in range(int(Nf/2)-1, -1, -1):
            if not above_HM[i]:
                break
            num_above_HM += 1

        # compute FWHM (Hz) using the number above HM and the freq difference
        FWHM = num_above_HM*(freqs[1] - freqs[0])

        return freqs, objs, FWHM

    def scan_power(self, probes=None, Ns=50, s_min=1e-2, s_max=1e2, solver='newton'):
        """ Scans the source amplitude and computes the objective function
            probes is a list of functions for computing the power, for example:
            [lambda simulation: simulation.flux_probe('x', [-NPML[0]-int(l/2/dl), ny + int(d/2/dl)], int(H/2/dl))]
        """

        if probes is None:
            raise ValueError("need to specify 'probes' kwarg as a list of functions for computing the power in each port.")
        num_probes = len(probes)

        # create src_amplitudes
        s_list = np.logspace(np.log10(s_min), np.log10(s_max), Ns)        

        bar = progressbar.ProgressBar(max_value=Ns)

        # transmission
        transmissions = [[] for _ in range(num_probes)]
        powers = []

        for i, s in enumerate(s_list):

            bar.update(i + 1)

            # make a new simulation object
            sim_new = copy.deepcopy(self.simulation)

            sim_new.modes[0].scale = s
            sim_new.modes[0].setup_src(sim_new)
            W_in = sim_new.W_in

            powers.append(W_in)

            if solver == 'hybrid':
                # NOTE: sometimes a mix of born and newtons method (hybrid) works best.  dont know why yet

                # compute the fields
                (_,_,_,c) = sim_new.solve_fields_nl(timing=False, averaging=False,
                            Estart=None, solver_nl='born', conv_threshold=1e-10,
                            max_num_iter=100)

                if c[-1] > 1e-10:
                    # compute the fields
                    (_,_,_,c) = sim_new.solve_fields_nl(timing=False, averaging=False,
                                Estart=None, solver_nl='newton', conv_threshold=1e-10,
                                max_num_iter=100)
            else:
                (_,_,_,c) = sim_new.solve_fields_nl(timing=False, averaging=False,
                            Estart=None, solver_nl=solver, conv_threshold=1e-10,
                            max_num_iter=100)

            # compute power transmission using each probe
            for probe_index, probe in enumerate(probes):
                W_out = probe(sim_new)
                transmissions[probe_index].append(W_out / W_in)

            # plt.show()

        return powers, transmissions

    def plot_transmissions(self, transmissions, legend=None):
        """ Plots the results of the power scan """

        for p in transmissions:
            plt.plot(p)
        plt.xscale('log')
        plt.xlabel('input source amplitude')
        plt.ylabel('power transmission')
        if legend is not None:
            plt.legend(legend)
        plt.show()

    def _check_temp_plt(self):
        if self.temp_plt is not None:
        # Clear the temp_plt folder from previous runs
            import os, shutil
            folder = self.temp_plt.folder
            if not os.path.exists(folder):
                os.makedirs(folder)
            for the_file in os.listdir(folder):
                file_path = os.path.join(folder, the_file)
                if os.path.isfile(file_path):
                    os.unlink(file_path)


import numpy as np
import copy

"""
This is where you can custom define functions that create relative permittivities
given a set of geometric parameters.
See three_port, two_port, ortho_port, and accelerator for exampels
"""

def get_grid(shape, dl):
    # computes the coordinates in the grid

    (Nx, Ny) = shape

    # coordinate vectors
    x_coord = np.linspace(-Nx/2*dl, Nx/2*dl, Nx)
    y_coord = np.linspace(-Ny/2*dl, Ny/2*dl, Ny)

    # x and y coordinate arrays
    xs, ys = np.meshgrid(x_coord, y_coord, indexing='ij')
    return (xs, ys)


def apply_regions(reg_list, xs, ys, eps_start):
    # feed this function a list of regions and some coordinates and it spits out a permittivity

    # if it's not a list, make it one
    if not isinstance(reg_list, list):
        reg_list = [reg_list]

    # initialize permittivity
    eps_r = np.ones(xs.shape)

    # loop through lambdas and apply masks
    for reg in reg_list:
        reg_vec = np.vectorize(reg)
        material_mask = reg_vec(xs, ys)
        eps_r[material_mask] = eps_start

    return eps_r


def three_port(L, H, w, d, l, spc, dl, NPML, eps_start):

    # CONSTRUCTS A ONE IN TWO OUT PORT DEVICE
    # L         : box length in L0
    # H         : box width  in L0
    # w         : waveguide widths in L0
    # d         : distance between out waveguide (centers) in L0
    # dl        : grid size in L0
    # shape     : shape of the permittivity output
    # eps_start : starting relative permittivity

    Nx = 2*NPML[0] + int((2*l + L)/dl)       # num. grids in horizontal
    Ny = 2*NPML[1] + int((H + 2*spc)/dl)   # num. grids in vertical
    nx, ny = int(Nx/2), int(Ny/2)            # halfway grid points
    shape = (Nx, Ny)                          # shape of domain (in num. grids)

    y_mid = dl*int(Ny/2-ny)
    y_bot = dl*int(Ny/2-ny-d/2/dl)
    y_top = dl*int(Ny/2-ny+d/2/dl)
    wg_width_px = int(w/dl)

    # x and y coordinate arrays
    xs, ys = get_grid(shape, dl)

    # define the regions
    box    = lambda x, y: (np.abs(x) < L/2) * (np.abs(y-y_mid) < H/2)
    wg_in  = lambda x, y: (x < 0)           * (np.abs(y-y_mid) < dl*wg_width_px/2)  # note, this slight offset is to fix gridding issues
    wg_top = lambda x, y: (x > 0)           * (np.abs(y-y_bot) < dl*wg_width_px/2)
    wg_bot = lambda x, y: (x > 0)           * (np.abs(y-y_top) < dl*wg_width_px/2)

    reg_list = [box, wg_in, wg_top, wg_bot]

    eps_r = apply_regions(reg_list, xs, ys, eps_start)
    design_region = apply_regions([box], xs, ys, 2) - 1

    return eps_r, design_region


def two_port(L, H, w, l, spc, dl, NPML, eps_start):

    # CONSTRUCTS A ONE IN ONE OUT PORT DEVICE
    # L         : design region length in L0
    # H         : design region width  in L0
    # w         : waveguide widths in L0
    # l         : distance between waveguide and design region in L0 (x)
    # spc       : spc bewtween PML and top/bottom of design region
    # dl        : grid size in L0
    # NPML      : number of PML grids in [x, y]
    # eps_start : starting relative permittivity

    Nx = 2*NPML[0] + int((2*l + L)/dl)       # num. grids in horizontal
    Ny = 2*NPML[1] + int((H + 2*spc)/dl)   # num. grids in vertical
    nx, ny = int(Nx/2), int(Ny/2)            # halfway grid points
    shape = (Nx, Ny)                          # shape of domain (in num. grids)

    # x and y coordinate arrays
    xs, ys = get_grid(shape, dl)

    # define regions
    box = lambda x, y: (np.abs(x) < L/2) * (np.abs(y) < H/2)
    wg  = lambda x, y: (np.abs(y) < w/2)

    eps_r = apply_regions([wg], xs, ys, eps_start=eps_start)
    design_region = apply_regions([box], xs, ys, eps_start=2) - 1

    return eps_r, design_region


def ortho_port(L, L2, H, H2, w, l, dl, NPML, eps_start):

    # CONSTRUCTS A TOP DOWN, LEFT RIGHT OUT PORT DEVICE
    # L         : waveguide design region length in L0
    # L2        : width of design region in L0
    # H         : waveguide design region height in L0
    # H2        : design region height in L0
    # w         : waveguide widths in L0
    # l         : distance between waveguide and design region in L0 (x)
    # spc       : spc bewtween PML and bottom of design region
    # dl        : grid size in L0
    # NPML      : number of PML grids in [x, y]
    # eps_start : starting relative permittivity

    Nx = 2*NPML[0] + int((2*l + L)/dl)       # num. grids in horizontal
    Ny = 2*NPML[1] + int((2*l + H)/dl)     # num. grids in vertical
    nx, ny = int(Nx/2), int(Ny/2)            # halfway grid points
    shape = (Nx, Ny)                         # shape of domain (in num. grids)

    # x and y coordinate arrays
    xs, ys = get_grid(shape, dl)

    # define regions\n",
    box  = lambda x, y: (np.abs(x) < L/2) * (np.abs(y) < w/2)
    box2 = lambda x, y: (np.abs(x) < L2/2) * (np.abs(y) < H2/2)
    box3 = lambda x, y: (-w/2 < x < w/2+dl/2) * (-H2/2 < y < H/2)
    wg   = lambda x, y: (np.abs(y) < w/2)
    wg2  = lambda x, y: (-w/2 < x < w/2+dl/2) * (y > 0)

    eps_r         = apply_regions([wg, wg2], xs, ys, eps_start=eps_start)
    design_region = apply_regions([box, box2, box3], xs, ys, eps_start=2)
    design_region = design_region - 1

    return eps_r, design_region


def accelerator(beta, gap, lambda0, L, spc, dl, NPML, eps_start):

    # CONSTRUCTS A periodic structure in y with two slabs surrounding a central gap
    # beta      : electron speed / speed of light
    # gap       : gap size in L0
    # lambda0   : free space wavelength in L0
    # L         : length of design region in L0
    # spc       : distance between PML and src.  src and design region
    # dl        : grid size in L0
    # NPML      : number of pml points in x and y
    # eps_start : starting relative permittivity

    Nx = 2*NPML[0] + int((4*spc + 2*L + gap)/dl)
    Ny = int(beta*lambda0/dl)
    nx, ny = int(Nx/2), int(Ny/2)            # halfway grid points
    shape = (Nx, Ny)                          # shape of domain (in num. grids)


    # x and y coordinate arrays
    xs, ys = get_grid(shape, dl)

    des = lambda x, y: (np.abs(x) > gap/2) * (np.abs(x) < gap/2 + L)

    eps_r = apply_regions([des], xs, ys, eps_start=eps_start)
    design_region = apply_regions([des], xs, ys, eps_start=2)
    design_region = design_region - 1

    return eps_r, design_region

def accelerator_multi(beta, gap, lambda0, L, spc, num_channels, dl, NPML, eps_start):

    # CONSTRUCTS A periodic structure in y with two slabs surrounding a central gap
    # beta      : electron speed / speed of light
    # gap       : gap size in L0
    # lambda0   : free space wavelength in L0
    # L         : length of design region in L0
    # spc       : distance between PML and src.  src and design region
    # dl        : grid size in L0
    # NPML      : number of pml points in x and y
    # eps_start : starting relative permittivity

    if num_channels % 2 == 1:
        Nx = 2*NPML[0] + int((4*spc + 2*L*num_channels + gap*num_channels)/dl)
    else:
        raise ValueError("must be odd number of channels")

    Ny = int(beta*lambda0/dl)
    nx, ny = int(Nx/2), int(Ny/2)            # halfway grid points
    shape = (Nx, Ny)                          # shape of domain (in num. grids)


    # x and y coordinate arrays
    xs, ys = get_grid(shape, dl)

    des_list = []
    centers = []
    for i in range(num_channels//2+1):
        centers.append(copy.deepcopy(i * (gap + L)))
        lambda_i = lambda x, y, i=i: (np.abs(x) - centers[i] > gap/2) * (np.abs(x) - centers[i] < gap/2 + L)
        des_list.append(lambda_i)

    print(centers)

    eps_r = apply_regions(des_list, xs, ys, eps_start=eps_start)
    design_region = apply_regions(des_list, xs, ys, eps_start=2)
    design_region = design_region - 1

    return eps_r, design_region    


import matplotlib.pylab as plt
import numpy as np
import autograd.numpy as npa
import scipy.sparse as sp
import progressbar

from autograd import grad
from functools import partial

def wrap(i, N):
    return i + (i >= N)*(i - N) + (i < 0)*(i + N)

def dist(r1, r2):
    return np.sqrt(np.sum(np.square(r1 - r2)))

def sub2ind(array_shape, rows, cols):
    return rows*array_shape[1] + cols

def get_W(Nx, Ny, design_region, NPML, R=10):

    diffs = range(-R, R+1)
    N = Nx*Ny

    # if there is no PML on one or more sides, need to wrap filter around
    is_periodic_x = NPML[0] == 0
    is_periodic_y = NPML[1] == 0

    row_indeces = []
    col_indeces = []
    vals = []

    if is_periodic_x:
        i_range = range(Nx)
    else:
        i_range = range(R + NPML[0] + 1, Nx - R - 1 - NPML[0])

    if is_periodic_y:
        j_range = range(Ny)
    else:
        j_range = range(R + NPML[1] + 1, Ny - R - 1 - NPML[1])

    bar = progressbar.ProgressBar(max_value=len(i_range)+1)

    for counti, i1 in enumerate(i_range):
        bar.update(counti)        
        for countj, j1 in enumerate(j_range):

            r1 = np.array([i1, j1])

            row_index = sub2ind((Nx, Ny), i1, j1)

            for i_diff in diffs:
                i2 = i1 + i_diff
                for j_diff in diffs:
                    j2 = j1 + j_diff
                    r2 = np.array([i2, j2])

                    # wrap around for periodic BC
                    i2 = wrap(i2, Nx)
                    j2 = wrap(j2, Ny)

                    col_index = sub2ind((Nx, Ny), i2, j2)

                    val = R - dist(r1, r2)

                    if val > 0:
                        row_indeces.append(row_index)
                        col_indeces.append(col_index)
                        vals.append(val)

    W = sp.csr_matrix((vals, (row_indeces, col_indeces)), shape=(N, N))

    des_vec = design_region.reshape((-1,))
    no_des_vec = 1-des_vec
    des_mat = sp.diags(des_vec, shape=(N, N))
    no_des_mat = sp.diags(no_des_vec, shape=(N, N))

    W_des = W.dot(des_mat) + no_des_mat

    norm_vec = W.dot(np.ones((Nx*Ny,)))
    norm_vec[norm_vec == 0] = 1
    norm_mat = sp.diags(1/norm_vec, shape=(N, N))

    W = W.dot(norm_mat)

    return des_mat.dot(W) + no_des_mat


""" THESE ARE THE OPERATIONS GOING FROM A DENSITY TO A PERMITTIVITY
    THROUGH FILTERING AND PROJECTING """


def rho2rhot(rho, W):
    # density to filtered density
    (Nx, Ny) = rho.shape
    rho_vec = rho.reshape((-1,))
    rhot_vec = W.dot(rho_vec)
    rhot = np.reshape(rhot_vec, (Nx, Ny))
    return rhot


def rhot2rhob(rhot, eta=0.5, beta=100):
    # filtered density to projected density
    num = np.tanh(beta*eta) + np.tanh(beta*(rhot - eta))
    den = np.tanh(beta*eta) + np.tanh(beta*(1 - eta))
    return num / den


def rhob2eps(rhob, eps_m):
    # filtered density to permittivity
    return 1 + rhob * (eps_m - 1)


def eps2rho(eps, eps_m):
    # permittivity to density (ONLY USED FOR STARTING THE SIMULATION!)
    return (eps - 1) / (eps_m - 1)


def rho2eps(rho, eps_m, W, eta=0.5, beta=100):
    rhot = rho2rhot(rho, W)
    rhob = rhot2rhob(rhot, eta=eta, beta=beta)
    eps = rhob2eps(rhob, eps_m=eps_m)
    return eps


"""" DERIVATIVE OPERATORS """


def drhot_drho(W):
    # derivative of filtered density with respect to design density
    return W


def drhob_drhot(rho_t, eta=0.5, beta=100):
    # change in projected density with respect to filtered density
    rhot_vec = np.reshape(rho_t, (-1,))
    num = beta - beta*np.square(np.tanh(beta*(rhot_vec - eta)))
    den = np.tanh(beta*eta) + np.tanh(beta*(1 - eta))
    return num / den

def deps_drhob(rhob, eps_m):
    return (eps_m - 1)



if __name__ == '__main__':

    ## NOTE: Running this file directly will generate some filtering plots used for testing
    # and as some figures in the supplementary information.

    """ some isolated tests of the filtering and projection """
    import matplotlib.style
    import matplotlib as mpl
    import matplotlib.gridspec as gridspec
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    mpl.style.use('seaborn-colorblind')    

    Nx = 100
    Ny = 100
    R = 5
    eps_m = 6
    beta = 100
    eta = 0.5
    NPML = [5, 5]

    rho = np.zeros((Nx, Ny))
    rho[20:Nx-20, Ny//2:Ny//2+20] = 0.7
    rho[Nx//2:Nx//2+5, 20:Ny-20] = 1
    # plt.imshow(rho)
    # plt.show()

    design_region = np.ones((Nx, Ny))

    W = get_W(Nx, Ny, design_region, NPML, R=R)

    rhot = rho2rhot(rho, W)
    rhob = rhot2rhob(rhot, eta=eta, beta=beta)
    eps = rhob2eps(rhob, eps_m=eps_m)

    f = plt.figure(figsize=(8, 4))
    RTs = np.linspace(0,1,1000)
    plt.plot(RTs, rhot2rhob(RTs, eta=0.5, beta=0.1))
    plt.plot(RTs, rhot2rhob(RTs, eta=0.5, beta=10))
    plt.plot(RTs, rhot2rhob(RTs, eta=0.5, beta=1000))
    plt.legend((r"$\beta = .1$", r"$\beta = 10$", r"$\beta = 1000$"))
    plt.xlabel(r"$\tilde{\rho}_i$")
    plt.ylabel(r"$\bar{\rho}_i$")
    plt.title(r"projection with varying $\beta$ ($\eta=0.5$)")
    # plt.savefig('data/figs/img/project.pdf', dpi=300)
    plt.show()

    # # change in projected density with respect to filtered density


    def circle(Nx, Ny, R=10):

        diffs = range(-R, R+1)
        N = Nx*Ny

        circ = np.zeros((Nx, Ny))

        for i1 in range(Nx//2, Nx//2+1):
            for j1 in range(Ny//2, Ny//2+1):
                r1 = np.array([i1, j1])

                if i1 <= R or i1 >= Nx-R-1:
                    pass
                elif j1 <= R or j1 >= Ny-R-1:
                    pass
                else:
                    for i_diff in diffs:
                        i2 = i1 + i_diff
                        for j_diff in diffs:
                            j2 = j1 + j_diff
                            r2 = np.array([i2, j2])

                            val = R - dist(r1, r2)

                            if val > 0:
                                circ[i2, j2] = val
        return circ

    f = plt.figure(figsize=(5, 5))    
    circ = circle(Nx, Ny, R=10)
    im = plt.imshow(circ, cmap='inferno')
    plt.colorbar()
    plt.title(r'filter response ($R = 10$ pixels)')
    plt.xlabel('pixels')
    plt.ylabel('pixels')
    # plt.savefig('data/figs/img/response.pdf', dpi=300)
    plt.show()

    def colorbar(mappable):
        ax = mappable.axes
        fig = ax.figure
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        return fig.colorbar(mappable, cax=cax)

    f = plt.figure(figsize=(9, 7), constrained_layout=True)
    gs = gridspec.GridSpec(2, 2, figure=f, height_ratios=[1, 1], wspace=0.1, hspace=0.5)
    ax_rho = plt.subplot(gs[0, 0])
    ax_rhot = plt.subplot(gs[0, 1])
    ax_rhob = plt.subplot(gs[1, 0])
    ax_eps = plt.subplot(gs[1, 1])

    im1 = ax_rho.imshow(rho, cmap='Greys')
    colorbar(im1)
    ax_rho.set_title(r'design density ($\rho$)')
    ax_rho.set_xlabel('pixels (x)')
    ax_rho.set_ylabel('pixels (y)')
    im2 = ax_rhot.imshow(rhot, cmap='Greys')
    colorbar(im2)
    ax_rhot.set_title(r'filtered density ($\tilde{\rho}$)')   
    ax_rhot.set_xlabel('pixels (x)')
    ax_rhot.set_ylabel('pixels (y)')
    im3 = ax_rhob.imshow(rhob, cmap='Greys')
    colorbar(im3)
    ax_rhob.set_title(r'projected density ($\bar{\rho}$)') 
    ax_rhob.set_xlabel('pixels (x)')
    ax_rhob.set_ylabel('pixels (y)')   
    im4 = ax_eps.imshow(eps, cmap='Greys')
    ax_eps.set_title(r'rel. permittivity ($\epsilon_r$)')
    ax_eps.set_xlabel('pixels (x)')
    ax_eps.set_ylabel('pixels (y)')
    colorbar(im4)

    # plt.savefig('data/figs/img/filter.pdf', dpi=300)
    plt.show()



import numpy as np
import scipy.sparse as sp

from angler.constants import ETA_0, EPSILON_0, DEFAULT_MATRIX_FORMAT


def sig_w(l, dw, m=4, lnR=-12):
    # helper for S()

    sig_max = -(m+1)*lnR/(2*ETA_0*dw)
    return sig_max*(l/dw)**m


def S(l, dw, omega, L0):
    # helper for create_sfactor()

    return 1 - 1j*sig_w(l, dw)/(omega*EPSILON_0*L0)


def create_sfactor(wrange, L0, s, omega, Nw, Nw_pml):
    # used to help construct the S matrices for the PML creation

    sfactor_array = np.ones(Nw, dtype=np.complex128)
    if Nw_pml < 1:
        return sfactor_array
    hw = np.diff(wrange)[0]/Nw
    dw = Nw_pml*hw
    for i in range(0, Nw):
        if s is 'f':
            if i <= Nw_pml:
                sfactor_array[i] = S(hw * (Nw_pml - i + 0.5), dw, omega, L0)
            elif i > Nw - Nw_pml:
                sfactor_array[i] = S(hw * (i - (Nw - Nw_pml) - 0.5), dw, omega, L0)
        if s is 'b':
            if i <= Nw_pml:
                sfactor_array[i] = S(hw * (Nw_pml - i + 1), dw, omega, L0)
            elif i > Nw - Nw_pml:
                sfactor_array[i] = S(hw * (i - (Nw - Nw_pml) - 1), dw, omega, L0)
    return sfactor_array


def S_create(omega, L0, N, Npml, xrange,
             yrange=None, matrix_format=DEFAULT_MATRIX_FORMAT):
    # creates S matrices for the PML creation

    M = np.prod(N)
    if np.isscalar(Npml):
        Npml = np.array([Npml])
    if len(N) < 2:
        N = np.append(N, 1)
        Npml = np.append(Npml, 0)
    Nx = N[0]
    Nx_pml = Npml[0]
    Ny = N[1]
    Ny_pml = Npml[1]

    # Create the sfactor in each direction and for 'f' and 'b'
    s_vector_x_f = create_sfactor(xrange, L0, 'f', omega, Nx, Nx_pml)
    s_vector_x_b = create_sfactor(xrange, L0, 'b', omega, Nx, Nx_pml)
    s_vector_y_f = create_sfactor(yrange, L0, 'f', omega, Ny, Ny_pml)
    s_vector_y_b = create_sfactor(yrange, L0, 'b', omega, Ny, Ny_pml)

    # Fill the 2D space with layers of appropriate s-factors
    Sx_f_2D = np.zeros(N, dtype=np.complex128)
    Sx_b_2D = np.zeros(N, dtype=np.complex128)
    Sy_f_2D = np.zeros(N, dtype=np.complex128)
    Sy_b_2D = np.zeros(N, dtype=np.complex128)

    for i in range(0, Ny):
        Sx_f_2D[:, i] = 1/s_vector_x_f
        Sx_b_2D[:, i] = 1/s_vector_x_b

    for i in range(0, Nx):
        Sy_f_2D[i, :] = 1/s_vector_y_f
        Sy_b_2D[i, :] = 1/s_vector_y_b

    # Reshape the 2D s-factors into a 1D s-array
    Sx_f_vec = Sx_f_2D.reshape((-1,))
    Sx_b_vec = Sx_b_2D.reshape((-1,))
    Sy_f_vec = Sy_f_2D.reshape((-1,))
    Sy_b_vec = Sy_b_2D.reshape((-1,))

    # Construct the 1D total s-array into a diagonal matrix
    Sx_f = sp.spdiags(Sx_f_vec, 0, M, M, format=matrix_format)
    Sx_b = sp.spdiags(Sx_b_vec, 0, M, M, format=matrix_format)
    Sy_f = sp.spdiags(Sy_f_vec, 0, M, M, format=matrix_format)
    Sy_b = sp.spdiags(Sy_b_vec, 0, M, M, format=matrix_format)

    return (Sx_f, Sx_b, Sy_f, Sy_b)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import LogNorm


def plt_base(field_val, outline_val, cmap, vmin, vmax, label,
             cbar=True, outline=None, ax=None, logscale=False):
    # Base plotting function for fields

    field_val = field_val.transpose()
    outline_val = outline_val.transpose()

    if ax is None:
        fig, ax = plt.subplots(1, constrained_layout=True)

    if not logscale:
        h = ax.imshow(field_val, cmap=cmap, vmin=vmin, vmax=vmax, origin='lower')
    else:
        h = ax.imshow(field_val, cmap=cmap, vmin=vmin, vmax=vmax, origin='lower', norm=LogNorm(vmin=vmin, vmax=vmax))

    if cbar:
        plt.colorbar(h, label=label, ax=ax)

    if outline:
        # Do black and white so we can see on both magma and RdBu
        ax.contour(outline_val, levels=2, linewidths=1.0, colors='w')
        ax.contour(outline_val, levels=2, linewidths=0.5, colors='k')

    ax.set_xticks([])
    ax.set_yticks([])

    return ax


def plt_base_eps(field_val, outline_val, cmap, vmin, vmax,
                 cbar=True, outline=None, ax=None):
    # Base plotting function for permittivity

    field_val = field_val.transpose()
    outline_val = outline_val.transpose()

    if ax is None:
        fig, ax = plt.subplots(1, constrained_layout=True)

    h = ax.imshow(field_val, cmap=cmap, vmin=vmin, vmax=vmax, origin='lower')

    if cbar:
        plt.colorbar(h, label='relative permittivity', ax=ax)

    if outline:
        # Do black and white so we can see on both magma and RdBu
        ax.contour(outline_val, levels=2, linewidths=1.0, colors='w')
        ax.contour(outline_val, levels=2, linewidths=0.5, colors='k')

    ax.set_xticks([])
    ax.set_yticks([])

    return ax


def plt_base_ani(field_val, cbar=True, Nframes=40, interval=80):

    field_val = field_val.transpose()

    fig, ax = plt.subplots(1, constrained_layout=True)
    h = ax.imshow(np.zeros(field_val.shape).transpose(), origin='lower')

    ax.set_xticks([])
    ax.set_yticks([])

    def init():
        vmax = np.abs(field_val).max()
        h.set_data(np.zeros(field_val.shape).transpose())
        h.set_cmap('RdBu')
        h.set_clim(vmin=-vmax, vmax=+vmax)

        return (h,)

    def animate(i):
        fields = np.real(field_val*np.exp(1j*2*np.pi*i/(Nframes-1)))
        h.set_data(fields)
        return (h,)

    plt.close()
    return mpl.animation.FuncAnimation(fig, animate, init_func=init,
                                       frames=Nframes, interval=interval,
                                       blit=True)

class Temp_plt():

    def __init__(self, it_plot=1, plot_what=('eps'), folder='figs/data/temp_im/', 
                    dpi=100, figsize=(4, 4), vlims=(None,None)):
        self.it_plot = it_plot
        self.plot_what = plot_what 
        self.folder = folder
        self.figsize = figsize
        self.dpi = dpi
        self.vlims = vlims


import numpy as np
import scipy.sparse as sp
import autograd.numpy as npa
from autograd import grad
from functools import partial

from angler.linalg import solver_direct, grid_average
from angler.derivatives import unpack_derivs
from angler.filter import deps_drhob, drhob_drhot
from angler.constants import *

def adjoint_linear_Ez(simulation, b_aj, averaging=False, solver=DEFAULT_SOLVER, matrix_format=DEFAULT_MATRIX_FORMAT):
    # Compute the adjoint field for a linear problem
    # Note: the correct definition requires simulating with the transpose matrix A.T
    EPSILON_0_ = EPSILON_0*simulation.L0
    MU_0_ = MU_0*simulation.L0
    omega = simulation.omega

    (Nx, Ny) = (simulation.Nx, simulation.Ny)
    M = Nx*Ny
    A = simulation.A

    ez = solver_direct(A.T, b_aj, solver=solver)
    Ez = ez.reshape((Nx, Ny))

    return Ez


def adjoint_linear_Hz(simulation, b_aj, averaging=False, solver=DEFAULT_SOLVER, matrix_format=DEFAULT_MATRIX_FORMAT):
    # Compute the adjoint field for a linear problem
    # Note: the correct definition requires simulating with the transpose matrix A.T
    EPSILON_0_ = EPSILON_0*simulation.L0
    MU_0_ = MU_0*simulation.L0
    omega = simulation.omega

    (Nx, Ny) = (simulation.Nx, simulation.Ny)
    M = Nx*Ny
    A = simulation.A

    hz = solver_direct(A.T, b_aj, solver=solver)
    (Dyb, Dxb, Dxf, Dyf) = unpack_derivs(simulation.derivs) 
    (Dyb_T, Dxb_T, Dxf_T, Dyf_T) = (Dyb.T, Dxb.T, Dxf.T, Dyf.T)

    if averaging:
        vector_eps_x = grid_average(EPSILON_0_*simulation.eps_r, 'x').reshape((-1,))
        vector_eps_y = grid_average(EPSILON_0_*simulation.eps_r, 'y').reshape((-1,))
    else:
        vector_eps_x = EPSILON_0_*simulation.eps_r.reshape((-1,))
        vector_eps_y = EPSILON_0_*simulation.eps_r.reshape((-1,))

    T_eps_x_inv = sp.spdiags(1/vector_eps_x, 0, M, M, format=matrix_format)
    T_eps_y_inv = sp.spdiags(1/vector_eps_y, 0, M, M, format=matrix_format)
    
    # Note: to get the correct gradient in the end, we must use Dxf, Dyf here   
    ex =  1/1j/omega * T_eps_y_inv.dot(Dyf_T).dot(hz).T
    ey = -1/1j/omega * T_eps_x_inv.dot(Dxf_T).dot(hz).T

    Ex = ex.reshape((Nx, Ny))
    Ey = ey.reshape((Nx, Ny))

    return (Ex, Ey)


def adjoint_kerr_Ez(simulation, b_aj,
                     averaging=False, solver=DEFAULT_SOLVER, matrix_format=DEFAULT_MATRIX_FORMAT):
    # Compute the adjoint field for a nonlinear problem
    # Note: written only for Ez!

    EPSILON_0_ = EPSILON_0*simulation.L0
    MU_0_ = MU_0*simulation.L0
    omega = simulation.omega

    (Nx, Ny) = (simulation.Nx, simulation.Ny)
    M = Nx*Ny

    Ez = simulation.fields_nl['Ez']
    Anl = simulation.A + simulation.Anl
    dAde = omega**2*EPSILON_0_*simulation.dnl_de

    C11 = Anl + sp.spdiags((dAde*Ez).reshape((-1,)), 0, M, M, format=matrix_format)
    C12 = sp.spdiags((np.conj(dAde)*Ez).reshape((-1)), 0, M, M, format=matrix_format)
    C_full = sp.vstack((sp.hstack((C11, C12)), np.conj(sp.hstack((C12, C11)))))
    b_aj = b_aj.reshape((-1,))

    ez = solver_direct(C_full.T, np.vstack((b_aj, np.conj(b_aj))), solver=solver)

    if np.linalg.norm(ez[range(M)] - np.conj(ez[range(M, 2*M)])) > 1e-8:
        print('Adjoint field and conjugate do not match; something might be wrong')

    Ez = ez[range(M)].reshape((Nx, Ny))

    return Ez



import numpy as np
import numpy.linalg as la
import scipy.sparse as sp
import scipy.sparse.linalg as spl
from copy import deepcopy

from scipy.optimize import newton_krylov, anderson

from angler.linalg import grid_average, solver_direct, solver_complex2real
from angler.derivatives import unpack_derivs
from angler.constants import (DEFAULT_LENGTH_SCALE, DEFAULT_MATRIX_FORMAT,
							  DEFAULT_SOLVER, EPSILON_0, MU_0)


def born_solve(simulation,
			   Estart=None, conv_threshold=1e-10, max_num_iter=50,
			   averaging=True):
	# solves for the nonlinear fields

	# Stores convergence parameters
	conv_array = np.zeros((max_num_iter, 1))

	# Defne the starting field for the simulation
	if Estart is None:
		if simulation.fields[simulation.pol] is None:
			# if its not specified, nor already, solve for the linear fields			
			(_, _, Fz) = simulation.solve_fields()
		else:
			# if its not specified, but stored, use that one
			Fz = deepcopy(simulation.fields[simulation.pol])
	else:
		# otherwise, use the specified version
		Fz = Estart

	# Solve iteratively
	for istep in range(max_num_iter):

		Fprev = Fz

		# set new permittivity
		simulation.compute_nl(Fprev)

		(Fx, Fy, Fz) = simulation.solve_fields(include_nl=True)

		# get convergence and break
		convergence = la.norm(Fz - Fprev)/la.norm(Fz)
		conv_array[istep] = convergence

		# if below threshold, break and return
		if convergence < conv_threshold:
			break

	if convergence > conv_threshold:
		print("the simulation did not converge, reached {}".format(convergence))

	return (Fx, Fy, Fz, conv_array)

def newton_solve(simulation,
				 Estart=None, conv_threshold=1e-10, max_num_iter=50,
				 averaging=True, solver=DEFAULT_SOLVER, jac_solver='c2r',
				 matrix_format=DEFAULT_MATRIX_FORMAT):
	# solves for the nonlinear fields using Newton's method

	# Stores convergence parameters
	conv_array = np.zeros((max_num_iter, 1))

	# num. columns and rows of A
	Nbig = simulation.Nx*simulation.Ny

	if simulation.pol == 'Ez':
		# Defne the starting field for the simulation
		if Estart is None:
			if simulation.fields['Ez'] is None:
				(_, _, Ez) = simulation.solve_fields()
			else:
				Ez = deepcopy(simulation.fields['Ez'])
		else:
			Ez = Estart

		# Solve iteratively
		for istep in range(max_num_iter):
			Eprev = Ez

			(fx, Jac11, Jac12) = nl_eq_and_jac(simulation, Ez=Eprev,
											   matrix_format=matrix_format)

			# Note: Newton's method is defined as a linear problem to avoid inverting the Jacobian
			# Namely, J*(x_n - x_{n-1}) = -f(x_{n-1}), where J = df/dx(x_{n-1})

			Ediff = solver_complex2real(Jac11, Jac12, fx,
										solver=solver, timing=False)
			# Abig = sp.sp_vstack((sp.sp_hstack((Jac11, Jac12)), \
			#   sp.sp_hstack((np.conj(Jac12), np.conj(Jac11)))))
			# Ediff = solver_direct(Abig, np.vstack((fx, np.conj(fx))))

			Ez = Eprev - Ediff[range(Nbig)].reshape(simulation.Nx, simulation.Ny)

			# get convergence and break
			convergence = la.norm(Ez - Eprev)/la.norm(Ez)
			conv_array[istep] = convergence

			# if below threshold, break and return
			if convergence < conv_threshold:
				break

		# Solve the fdfd problem with the final eps_nl
		simulation.compute_nl(Ez)
		(Hx, Hy, Ez) = simulation.solve_fields(include_nl=True)

		if convergence > conv_threshold:
			print("the simulation did not converge, reached {}".format(convergence))

		return (Hx, Hy, Ez, conv_array)

	else:
		raise ValueError('Invalid polarization: {}'.format(str(self.pol)))

def nl_eq_and_jac(simulation,
				  averaging=True, Ex=None, Ey=None, Ez=None, compute_jac=True,
				  matrix_format=DEFAULT_MATRIX_FORMAT):
	# Evaluates the nonlinear function f(E) that defines the problem to solve f(E) = 0, as well as the Jacobian df/dE
	# Could add a check that only Ez is None for Hz polarization and vice-versa

	omega = simulation.omega
	EPSILON_0_ = EPSILON_0*simulation.L0
	MU_0_ = MU_0*simulation.L0

	Nbig = simulation.Nx*simulation.Ny

	if simulation.pol == 'Ez':
		simulation.compute_nl(Ez)
		Anl = simulation.A + simulation.Anl
		fE = (Anl.dot(Ez.reshape(-1,)) - simulation.src.reshape(-1,)*1j*omega)

		# Make it explicitly a column vector
		fE = fE.reshape(-1,)

		if compute_jac:
			simulation.compute_nl(Ez)
			dAde = (simulation.dnl_de).reshape((-1,))*omega**2*EPSILON_0_
			Jac11 = Anl + sp.spdiags(dAde*Ez.reshape((-1,)), 0, Nbig, Nbig, format=matrix_format)
			Jac12 = sp.spdiags(np.conj(dAde)*Ez.reshape((-1,)), 0, Nbig, Nbig, format=matrix_format)

	elif simulation.pol == 'Hz':
		raise ValueError('angler doesnt support newton method for Hz polarization yet')		

	else:
		raise ValueError('Invalid polarization: {}'.format(str(self.pol)))

	if compute_jac:
		return (fE, Jac11, Jac12)
	else:
		return fE


def newton_krylov_solve(simulation, Estart=None, conv_threshold=1e-10, max_num_iter=50,
				 averaging=True, solver=DEFAULT_SOLVER, jac_solver='c2r',
				 matrix_format=DEFAULT_MATRIX_FORMAT):
	# THIS DOESNT WORK YET! 

	# Stores convergence parameters
	conv_array = np.zeros((max_num_iter, 1))

	# num. columns and rows of A
	Nbig = simulation.Nx*simulation.Ny

	# Defne the starting field for the simulation
	if Estart is None:
		if simulation.fields['Ez'] is None:
			(_, _, E0) = simulation.solve_fields()
		else:
			E0 = deepcopy(simulation.fields['Ez'])
	else:
		E0 = Estart

	E0 = np.reshape(E0, (-1,))

	# funtion for roots
	def _f(E):
		E = np.reshape(E, simulation.eps_r.shape)
		fx = nl_eq_and_jac(simulation, Ez=E, compute_jac=False, matrix_format=matrix_format)
		return np.reshape(fx, (-1,))

	print(_f(E0))

	E_nl = newton_krylov(_f, E0, verbose=True)

	# I'm returning these in place of Hx and Hy to not break things
	return (E_nl, E_nl, E_nl, conv_array)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()

dependencies = [
        'pyMKL',
        'numpy',
        'scipy',
        'matplotlib>=2.2.2',
        'progressbar2==3.37.1',
        'autograd',
        'future'
]

setup(
    name='angler',
    version='0.0.14',
    description='Adjoint Nonlinear Gradients',
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Tyler Hughes, Momchil Minkov, Ian Williamson',
    author_email='tylerwhughes91@gmail.com',
    url='https://github.com/fancompute/angler',
    packages=find_packages(),
    install_requires=dependencies,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import numpy as np
import scipy.sparse as sp
from inspect import signature

from angler.adjoint import adjoint_linear_Ez, adjoint_kerr_Ez, adjoint_linear_Hz
from angler.filter import (eps2rho, rho2eps, get_W, deps_drhob, drhob_drhot,
                    drhot_drho, rho2rhot, drhot_drho, rhot2rhob)
from angler.constants import *
from angler.derivatives import unpack_derivs
from angler.linalg import grid_average

""" This is where the gradients are defined
    These are selected when you define an objective function
"""

def grad_linear_Ez(optimization, dJ, Ez, args):
    """gives the linear field gradient: partial J/ partial * E_lin dE_lin / deps"""

    b_aj = -dJ(*args)
    Ez_aj = adjoint_linear_Ez(optimization.simulation, b_aj)

    EPSILON_0_ = EPSILON_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    dAdeps = optimization.design_region*omega**2*EPSILON_0_

    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    Ez_vec = np.reshape(Ez, (-1,))

    dAdeps_vec = np.reshape(dAdeps, (-1,))
    dfdrho = eps_mat*filt_mat.multiply(Ez_vec*proj_mat*dAdeps_vec)
    Ez_aj_vec = np.reshape(Ez_aj, (-1,))
    sensitivity_vec = dfdrho.dot(Ez_aj_vec)        

    return 1*np.real(np.reshape(sensitivity_vec, rho.shape))

def grad_linear_Hx(optimization, dJ, Ez, args):
    """gives the linear field gradient: partial J/ partial * E_lin dE_lin / deps"""

    # get the adjoint Ez corresponding to Hx
    Dyb = optimization.simulation.derivs['Dyb']
    partial = -dJ(*args)
    partial_vec = partial.reshape((-1,))

    MU_0_ = MU_0*optimization.simulation.L0
    omega = optimization.simulation.omega

    b_aj_vec = -1/1j/omega/MU_0_ * Dyb.T.dot(partial_vec)
    b_aj = b_aj_vec.reshape(Ez.shape)

    # rest is the same
    Ez_aj = adjoint_linear_Ez(optimization.simulation, b_aj)

    EPSILON_0_ = EPSILON_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    dAdeps = optimization.design_region*omega**2*EPSILON_0_

    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    Ez_vec = np.reshape(Ez, (-1,))

    dAdeps_vec = np.reshape(dAdeps, (-1,))
    dfdrho = eps_mat*filt_mat.multiply(Ez_vec*proj_mat*dAdeps_vec)
    Ez_aj_vec = np.reshape(Ez_aj, (-1,))
    sensitivity_vec = dfdrho.dot(Ez_aj_vec)        

    return 1*np.real(np.reshape(sensitivity_vec, rho.shape))

def grad_linear_Hy(optimization, dJ, Ez, args):
    """gives the linear field gradient: partial J/ partial * E_lin dE_lin / deps"""

    # get the adjoint Ez corresponding to Hx
    Dxb = optimization.simulation.derivs['Dxb']
    partial = -dJ(*args)
    partial_vec = partial.reshape((-1,))

    MU_0_ = MU_0*optimization.simulation.L0
    omega = optimization.simulation.omega

    b_aj_vec = 1/1j/omega/MU_0_ * Dxb.T.dot(partial_vec)
    b_aj = b_aj_vec.reshape(Ez.shape)

    # rest is the same
    Ez_aj = adjoint_linear_Ez(optimization.simulation, b_aj)

    EPSILON_0_ = EPSILON_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    dAdeps = optimization.design_region*omega**2*EPSILON_0_

    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    Ez_vec = np.reshape(Ez, (-1,))

    dAdeps_vec = np.reshape(dAdeps, (-1,))
    dfdrho = eps_mat*filt_mat.multiply(Ez_vec*proj_mat*dAdeps_vec)
    Ez_aj_vec = np.reshape(Ez_aj, (-1,))
    sensitivity_vec = dfdrho.dot(Ez_aj_vec)        

    return 1*np.real(np.reshape(sensitivity_vec, rho.shape))

def grad_kerr_Ez(optimization, dJ, Ez_nl, args):
    """gives the linear field gradient: partial J/ partial * E_lin dE_lin / deps"""

    b_aj = -dJ(*args)
    Ez_aj = adjoint_kerr_Ez(optimization.simulation, b_aj)
    optimization.simulation.compute_nl(Ez_nl)

    EPSILON_0_ = EPSILON_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    dAdeps = optimization.design_region*omega**2*EPSILON_0_
    dAnldeps = dAdeps + optimization.design_region*omega**2*EPSILON_0_*optimization.simulation.dnl_deps

    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    Ez_vec = np.reshape(Ez_nl, (-1,))

    dfdrho = eps_mat*filt_mat.multiply(Ez_vec*proj_mat*np.reshape(dAnldeps, (-1,)))
    return 1*np.real(np.reshape(dfdrho.dot(np.reshape(Ez_aj, (-1,))), rho.shape))


def grad_kerr_Hx(optimization, dJ, Ez_nl, args):

    # get the adjoint source corresponding to Hx
    Dyb = optimization.simulation.derivs['Dyb']
    partial = -dJ(*args)
    partial_vec = partial.reshape((-1,))

    MU_0_ = MU_0*optimization.simulation.L0
    omega = optimization.simulation.omega

    b_aj_vec = -1/1j/omega/MU_0_ * Dyb.T.dot(partial_vec)
    b_aj = b_aj_vec.reshape(Ez.shape)

    # everything else is the same
    Ez_aj = adjoint_kerr_Ez(optimization.simulation, b_aj)
    optimization.simulation.compute_nl(Ez_nl)

    EPSILON_0_ = EPSILON_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    dAdeps = optimization.design_region*omega**2*EPSILON_0_
    dAnldeps = dAdeps + optimization.design_region*omega**2*EPSILON_0_*optimization.simulation.dnl_deps

    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    Ez_vec = np.reshape(Ez_nl, (-1,))

    dfdrho = eps_mat*filt_mat.multiply(Ez_vec*proj_mat*np.reshape(dAnldeps, (-1,)))
    return 1*np.real(np.reshape(dfdrho.dot(np.reshape(Ez_aj, (-1,))), rho.shape))



def grad_kerr_Hy(optimization, dJ, Ez_nl, args):
    """gives the linear field gradient: partial J/ partial * E_lin dE_lin / deps"""

    # get the adjoint source corresponding to Hy
    Dxb = optimization.simulation.derivs['Dxb']
    partial = -dJ(*args)
    partial_vec = partial.reshape((-1,))

    MU_0_ = MU_0*optimization.simulation.L0
    omega = optimization.simulation.omega

    b_aj_vec = 1/1j/omega/MU_0_ * Dxb.T.dot(partial_vec)
    b_aj = b_aj_vec.reshape(Ez.shape)

    # everything else is the same
    Ez_aj = adjoint_kerr_Ez(optimization.simulation, b_aj)
    optimization.simulation.compute_nl(Ez_nl)

    EPSILON_0_ = EPSILON_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    dAdeps = optimization.design_region*omega**2*EPSILON_0_
    dAnldeps = dAdeps + optimization.design_region*omega**2*EPSILON_0_*optimization.simulation.dnl_deps

    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    Ez_vec = np.reshape(Ez_nl, (-1,))

    dfdrho = eps_mat*filt_mat.multiply(Ez_vec*proj_mat*np.reshape(dAnldeps, (-1,)))
    return 1*np.real(np.reshape(dfdrho.dot(np.reshape(Ez_aj, (-1,))), rho.shape))




####################################################################################################################################
####################################################################################################################################
####################################################################################################################################
####################################################################################################################################
####################################################################################################################################


def grad_linear_Hz(optimization, dJ, Hz, args, averaging=False):

    EPSILON_0_ = EPSILON_0 * optimization.simulation.L0
    MU_0_ = MU_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    Dxb = optimization.simulation.derivs['Dxf']
    Dyb = optimization.simulation.derivs['Dyf']

    eps_tot = optimization.simulation.eps_r
    M = eps_tot.size

    if averaging:
        eps_x = grid_average(EPSILON_0_*(eps_tot), 'x')
        vector_eps_x = eps_x.reshape((-1,))
        eps_y = grid_average(EPSILON_0_*(eps_tot), 'y')
        vector_eps_y = eps_y.reshape((-1,))
    else:
        vector_eps_x = EPSILON_0_*(eps_tot).reshape((-1,))
        vector_eps_y = EPSILON_0_*(eps_tot).reshape((-1,))

    T_eps_x_inv = sp.spdiags(1/vector_eps_x, 0, M, M,
                          format=DEFAULT_MATRIX_FORMAT)
    T_eps_y_inv = sp.spdiags(1/vector_eps_y, 0, M, M,
                          format=DEFAULT_MATRIX_FORMAT)

    # get fields
    Hz_vec = np.reshape(Hz, (-1,))
    Ex_vec =  1/1j/omega * T_eps_y_inv.dot(Dyb).dot(Hz_vec)
    Ey_vec = -1/1j/omega * T_eps_x_inv.dot(Dxb).dot(Hz_vec) 

    # adjoint
    # partial = -dJ(*args)
    # partial_vec = partial.reshape((-1,))

    b_aj = -dJ(*args)
    (Ex_aj, Ey_aj) = adjoint_linear_Hz(optimization.simulation, b_aj, averaging=averaging)   
    Ex_aj_vec = np.reshape(Ex_aj, (-1,)).T
    Ey_aj_vec = np.reshape(Ey_aj, (-1,)).T

    # set up filtering bits
    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    if averaging:

        design_region_x = grid_average(optimization.design_region, 'x')
        dAdeps_x = design_region_x*omega**2*EPSILON_0_            
        design_region_y = grid_average(optimization.design_region, 'y')
        dAdeps_y = design_region_y*omega**2*EPSILON_0_
        dAdeps_vec_x = np.reshape(dAdeps_x, (-1,))
        dAdeps_vec_y = np.reshape(dAdeps_y, (-1,))
        dfdrho_x = eps_mat*filt_mat.multiply(Ex_vec*proj_mat*dAdeps_vec_x)
        dfdrho_y = eps_mat*filt_mat.multiply(Ey_vec*proj_mat*dAdeps_vec_x)
        sensitivity_vec = dfdrho_x.dot(Ex_aj_vec) + dfdrho_y.dot(Ey_aj_vec)    

    else:

        dAdeps = optimization.design_region*omega**2*EPSILON_0_    # Note: physical constants go here if need be!
        dAdeps_vec = np.reshape(dAdeps, (-1,))
        dfdrho_x = eps_mat*filt_mat.multiply(Ex_vec*proj_mat*dAdeps_vec)
        dfdrho_y = eps_mat*filt_mat.multiply(Ey_vec*proj_mat*dAdeps_vec)
        sensitivity_vec = dfdrho_x.dot(Ex_aj_vec) + dfdrho_y.dot(Ey_aj_vec)    

    return 1*np.real(np.reshape(sensitivity_vec, rho.shape))    

def grad_linear_Ex(optimization, dJ, Hz, args, averaging=False):

    EPSILON_0_ = EPSILON_0 * optimization.simulation.L0
    MU_0_ = MU_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    Dxb = optimization.simulation.derivs['Dxf']
    Dyb = optimization.simulation.derivs['Dyf']

    eps_tot = optimization.simulation.eps_r
    M = eps_tot.size

    if averaging:
        eps_x = grid_average(EPSILON_0_*(eps_tot), 'x')
        vector_eps_x = eps_x.reshape((-1,))
        eps_y = grid_average(EPSILON_0_*(eps_tot), 'y')
        vector_eps_y = eps_y.reshape((-1,))
    else:
        vector_eps_x = EPSILON_0_*(eps_tot).reshape((-1,))
        vector_eps_y = EPSILON_0_*(eps_tot).reshape((-1,))

    T_eps_x_inv = sp.spdiags(1/vector_eps_x, 0, M, M,
                          format=DEFAULT_MATRIX_FORMAT)
    T_eps_y_inv = sp.spdiags(1/vector_eps_y, 0, M, M,
                          format=DEFAULT_MATRIX_FORMAT)

    # get fields
    Hz_vec = np.reshape(Hz, (-1,))
    Ex_vec =  1/1j/omega * T_eps_y_inv.dot(Dyb).dot(Hz_vec)
    Ey_vec = -1/1j/omega * T_eps_x_inv.dot(Dxb).dot(Hz_vec) 

    partial = -dJ(*args)
    partial_vec = partial.reshape((-1,))

    b_aj_vec =  1/1j/omega * T_eps_y_inv.dot(Dyb.T).dot(partial_vec)
    b_aj = b_aj_vec.reshape(Hz.shape)

    (Ex_aj, Ey_aj) = adjoint_linear_Hz(optimization.simulation, b_aj, averaging=averaging)   
    Ex_aj_vec = np.reshape(Ex_aj, (-1,)).T
    Ey_aj_vec = np.reshape(Ey_aj, (-1,)).T

    # set up filtering bits
    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    if averaging:

        design_region_x = grid_average(optimization.design_region, 'x')
        dAdeps_x = design_region_x*omega**2*EPSILON_0_            
        design_region_y = grid_average(optimization.design_region, 'y')
        dAdeps_y = design_region_y*omega**2*EPSILON_0_
        dAdeps_vec_x = np.reshape(dAdeps_x, (-1,))
        dAdeps_vec_y = np.reshape(dAdeps_y, (-1,))
        dfdrho_x = eps_mat*filt_mat.multiply(Ex_vec*proj_mat*dAdeps_vec_x)
        dfdrho_y = eps_mat*filt_mat.multiply(Ey_vec*proj_mat*dAdeps_vec_x)
        sensitivity_vec = dfdrho_x.dot(Ex_aj_vec) + dfdrho_y.dot(Ey_aj_vec)    

    else:

        dAdeps = optimization.design_region*omega**2*EPSILON_0_    # Note: physical constants go here if need be!
        dAdeps_vec = np.reshape(dAdeps, (-1,))
        dfdrho_x = eps_mat*filt_mat.multiply(Ex_vec*proj_mat*dAdeps_vec)
        dfdrho_y = eps_mat*filt_mat.multiply(Ey_vec*proj_mat*dAdeps_vec)
        sensitivity_vec = dfdrho_x.dot(Ex_aj_vec) + dfdrho_y.dot(Ey_aj_vec)    

    return 1*np.real(np.reshape(sensitivity_vec, rho.shape)) 

def grad_linear_Ey(optimization, dJ, Hz, args, averaging=False):

    EPSILON_0_ = EPSILON_0 * optimization.simulation.L0
    MU_0_ = MU_0*optimization.simulation.L0
    omega = optimization.simulation.omega
    Dxb = optimization.simulation.derivs['Dxf']
    Dyb = optimization.simulation.derivs['Dyf']

    eps_tot = optimization.simulation.eps_r
    M = eps_tot.size

    if averaging:
        eps_x = grid_average(EPSILON_0_*(eps_tot), 'x')
        vector_eps_x = eps_x.reshape((-1,))
        eps_y = grid_average(EPSILON_0_*(eps_tot), 'y')
        vector_eps_y = eps_y.reshape((-1,))
    else:
        vector_eps_x = EPSILON_0_*(eps_tot).reshape((-1,))
        vector_eps_y = EPSILON_0_*(eps_tot).reshape((-1,))

    T_eps_x_inv = sp.spdiags(1/vector_eps_x, 0, M, M,
                          format=DEFAULT_MATRIX_FORMAT)
    T_eps_y_inv = sp.spdiags(1/vector_eps_y, 0, M, M,
                          format=DEFAULT_MATRIX_FORMAT)

    # get fields
    Hz_vec = np.reshape(Hz, (-1,))
    Ex_vec =  1/1j/omega * T_eps_y_inv.dot(Dyb).dot(Hz_vec)
    Ey_vec = -1/1j/omega * T_eps_x_inv.dot(Dxb).dot(Hz_vec) 

    partial = -dJ(*args)
    partial_vec = partial.reshape((-1,))

    b_aj_vec = -1/1j/omega * T_eps_x_inv.dot(Dxb.T).dot(partial_vec)
    b_aj = b_aj_vec.reshape(Hz.shape)

    (Ex_aj, Ey_aj) = adjoint_linear_Hz(optimization.simulation, b_aj, averaging=averaging)   
    Ex_aj_vec = np.reshape(Ex_aj, (-1,)).T
    Ey_aj_vec = np.reshape(Ey_aj, (-1,)).T

    # set up filtering bits
    rho = optimization.simulation.rho
    rho_t = rho2rhot(rho, optimization.W)
    rho_b = rhot2rhob(rho_t, eta=optimization.eta, beta=optimization.beta)
    eps_mat = (optimization.eps_m - 1)

    filt_mat = drhot_drho(optimization.W)
    proj_mat = drhob_drhot(rho_t, eta=optimization.eta, beta=optimization.beta)

    if averaging:

        design_region_x = grid_average(optimization.design_region, 'x')
        dAdeps_x = design_region_x*omega**2*EPSILON_0_            
        design_region_y = grid_average(optimization.design_region, 'y')
        dAdeps_y = design_region_y*omega**2*EPSILON_0_
        dAdeps_vec_x = np.reshape(dAdeps_x, (-1,))
        dAdeps_vec_y = np.reshape(dAdeps_y, (-1,))
        dfdrho_x = eps_mat*filt_mat.multiply(Ex_vec*proj_mat*dAdeps_vec_x)
        dfdrho_y = eps_mat*filt_mat.multiply(Ey_vec*proj_mat*dAdeps_vec_x)
        sensitivity_vec = dfdrho_x.dot(Ex_aj_vec) + dfdrho_y.dot(Ey_aj_vec)    

    else:

        dAdeps = optimization.design_region*omega**2*EPSILON_0_    # Note: physical constants go here if need be!
        dAdeps_vec = np.reshape(dAdeps, (-1,))
        dfdrho_x = eps_mat*filt_mat.multiply(Ex_vec*proj_mat*dAdeps_vec)
        dfdrho_y = eps_mat*filt_mat.multiply(Ey_vec*proj_mat*dAdeps_vec)
        sensitivity_vec = dfdrho_x.dot(Ex_aj_vec) + dfdrho_y.dot(Ey_aj_vec)    

    return 1*np.real(np.reshape(sensitivity_vec, rho.shape))

def grad_kerr_Hz(optimization, dJ, Ez_nl, args):
    raise NotImplementedError("need to write gradient for kerr Hz")

def grad_kerr_Ex(optimization, dJ, Ez_nl, args):
    raise NotImplementedError("need to write gradient for kerr Ex")

def grad_kerr_Ey(optimization, dJ, Ez_nl, args):
    raise NotImplementedError("need to write gradient for kerr Ey")

